# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the Transactions class of the tac negotiation skill."""

import datetime
import logging
from pathlib import Path
from typing import Deque, Tuple, cast
from unittest.mock import Mock, patch

import pytest

from aea.decision_maker.gop import GoalPursuitReadiness, OwnershipState, Preferences
from aea.exceptions import AEAEnforceError
from aea.helpers.transaction.base import Terms
from aea.protocols.dialogue.base import DialogueLabel
from aea.test_tools.test_skill import BaseSkillTestCase, COUNTERPARTY_AGENT_ADDRESS

from packages.fetchai.skills.tac_negotiation.dialogues import FipaDialogue
from packages.fetchai.skills.tac_negotiation.transactions import Transactions

from tests.conftest import ROOT_DIR


class TestTransactions(BaseSkillTestCase):
    """Test Transactions class of tac negotiation."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_negotiation")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        tac_dm_context_kwargs = {
            "goal_pursuit_readiness": GoalPursuitReadiness(),
            "ownership_state": OwnershipState(),
            "preferences": Preferences(),
        }
        super().setup(dm_context_kwargs=tac_dm_context_kwargs)
        cls.pending_transaction_timeout = 30
        cls.transactions = Transactions(
            pending_transaction_timeout=cls.pending_transaction_timeout,
            name="transactions",
            skill_context=cls._skill.skill_context,
        )

        cls.nonce = "125"
        cls.sender = "some_sender_address"
        cls.counterparty = "some_counterparty_address"
        cls.ledger_id = "some_ledger_id"
        cls.terms = Terms(
            ledger_id=cls.ledger_id,
            sender_address=cls.sender,
            counterparty_address=cls.counterparty,
            amount_by_currency_id={"1": 10},
            quantities_by_good_id={"2": -5},
            is_sender_payable_tx_fee=True,
            nonce=cls.nonce,
            fee_by_currency_id={"1": 1},
        )
        cls.dialogue_label = DialogueLabel(
            ("", ""),
            COUNTERPARTY_AGENT_ADDRESS,
            cls._skill.skill_context.agent_address,
        )
        cls.proposal_id = 5
        cls.transaction_id = "some_transaction_id"

    def test_simple_properties(self):
        """Test the properties of Transactions class."""
        assert self.transactions.pending_proposals == {}
        assert self.transactions.pending_initial_acceptances == {}

    def test_get_next_nonce(self):
        """Test the get_next_nonce method of the Transactions class."""
        assert self.transactions.get_next_nonce() == "1"
        assert self.transactions._nonce == 1

    def test_update_confirmed_transactions(self):
        """Test the update_confirmed_transactions method of the Transactions class."""
        # setup
        self.skill.skill_context._get_agent_context().shared_state[
            "confirmed_tx_ids"
        ] = [self.transaction_id]
        self.transactions._locked_txs[self.transaction_id] = self.terms
        self.transactions._locked_txs_as_buyer[self.transaction_id] = self.terms
        self.transactions._locked_txs_as_seller[self.transaction_id] = self.terms

        # operation
        self.transactions.update_confirmed_transactions()

        # after
        assert self.transactions._locked_txs == {}
        assert self.transactions._locked_txs_as_buyer == {}
        assert self.transactions._locked_txs_as_seller == {}

    def test_cleanup_pending_transactions_i(self):
        """Test the cleanup_pending_transactions method of the Transactions class where _last_update_for_transactions is NOT empty."""
        # setup
        datetime_mock = Mock(wraps=datetime.datetime)
        datetime_mock.now.return_value = datetime.datetime.strptime(
            "01 01 2020  00:03", "%d %m %Y %H:%M"
        )
        with patch("datetime.datetime", new=datetime_mock):
            self.transactions._register_transaction_with_time(self.transaction_id)

        self.transactions._locked_txs[self.transaction_id] = self.terms
        self.transactions._locked_txs_as_buyer[self.transaction_id] = self.terms
        self.transactions._locked_txs_as_seller[self.transaction_id] = self.terms

        # operation
        with patch.object(self.skill.skill_context.logger, "log") as mock_logger:
            self.transactions.cleanup_pending_transactions()

        # after
        mock_logger.assert_any_call(
            logging.DEBUG,
            f"removing transaction from pending list: {self.transaction_id}",
        )
        assert self.transactions._locked_txs == {}
        assert self.transactions._locked_txs_as_buyer == {}
        assert self.transactions._locked_txs_as_seller == {}

    def test_cleanup_pending_transactions_ii(self):
        """Test the cleanup_pending_transactions method of the Transactions class where _last_update_for_transactions is empty."""
        # setup
        cast(
            Deque[Tuple[datetime.datetime, str]],
            self.transactions._last_update_for_transactions,
        )

        assert self.transactions._locked_txs == {}
        assert self.transactions._locked_txs_as_buyer == {}
        assert self.transactions._locked_txs_as_seller == {}

        # operation
        self.transactions.cleanup_pending_transactions()

        # after
        assert self.transactions._locked_txs == {}
        assert self.transactions._locked_txs_as_buyer == {}
        assert self.transactions._locked_txs_as_seller == {}

    def test_add_pending_proposal_i(self):
        """Test the add_pending_proposal method of the Transactions class."""
        # before
        assert self.dialogue_label not in self.transactions._pending_proposals

        # operation
        self.transactions.add_pending_proposal(
            self.dialogue_label, self.proposal_id, self.terms
        )

        # after
        assert (
            self.transactions._pending_proposals[self.dialogue_label][self.proposal_id]
            == self.terms
        )

    def test_add_pending_proposal_ii(self):
        """Test the add_pending_proposal method of the Transactions class where dialogue_label IS in _pending_proposals."""
        # setup
        self.transactions._pending_proposals[self.dialogue_label] = {1: self.terms}

        # operation
        with pytest.raises(
            AEAEnforceError,
            match="Proposal is already in the list of pending proposals.",
        ):
            self.transactions.add_pending_proposal(
                self.dialogue_label, self.proposal_id, self.terms
            )

    def test_add_pending_proposal_iii(self):
        """Test the add_pending_proposal method of the Transactions class where proposal_id IS in _pending_proposals."""
        # setup
        self.transactions._pending_proposals[self.dialogue_label][
            self.proposal_id
        ] = self.terms

        # operation
        with pytest.raises(
            AEAEnforceError,
            match="Proposal is already in the list of pending proposals.",
        ):
            self.transactions.add_pending_proposal(
                self.dialogue_label, self.proposal_id, self.terms
            )

    def test_pop_pending_proposal_i(self):
        """Test the pop_pending_proposal method of the Transactions class."""
        # setup
        self.transactions.add_pending_proposal(
            self.dialogue_label, self.proposal_id, self.terms
        )

        # operation
        actual_terms = self.transactions.pop_pending_proposal(
            self.dialogue_label, self.proposal_id
        )

        # after
        assert actual_terms == self.terms
        assert (
            self.proposal_id
            not in self.transactions._pending_proposals[self.dialogue_label]
        )

    def test_pop_pending_proposal_ii(self):
        """Test the pop_pending_proposal method of the Transactions class where dialogue_label IS in _pending_proposals."""
        # setup
        self.transactions.add_pending_proposal(
            self.dialogue_label, self.proposal_id, self.terms
        )
        self.transactions._pending_proposals = {}

        # operation
        with pytest.raises(
            AEAEnforceError,
            match="Cannot find the proposal in the list of pending proposals.",
        ):
            assert self.transactions.pop_pending_proposal(
                self.dialogue_label, self.proposal_id
            )

    def test_pop_pending_proposal_iii(self):
        """Test the pop_pending_proposal method of the Transactions class where dialogue_label and proposal_id IS in _pending_proposals."""
        # setup
        self.transactions.add_pending_proposal(
            self.dialogue_label, self.proposal_id, self.terms
        )
        self.transactions._pending_proposals[self.dialogue_label] = {1: self.terms}

        # operation
        with pytest.raises(
            AEAEnforceError,
            match="Cannot find the proposal in the list of pending proposals.",
        ):
            assert self.transactions.pop_pending_proposal(
                self.dialogue_label, self.proposal_id
            )

    def test_add_pending_initial_acceptance_i(self):
        """Test the add_pending_initial_acceptance method of the Transactions class."""
        # before
        assert self.transactions._pending_initial_acceptances == {}

        # operation
        self.transactions.add_pending_initial_acceptance(
            self.dialogue_label, self.proposal_id, self.terms,
        )

        # after
        assert (
            self.transactions._pending_initial_acceptances[self.dialogue_label][
                self.proposal_id
            ]
            == self.terms
        )

    def test_add_pending_initial_acceptance_ii(self):
        """Test the add_pending_initial_acceptance method of the Transactions class where dialogue_label IS in _pending_initial_acceptances."""
        # setup
        self.transactions._pending_initial_acceptances[self.dialogue_label] = {
            1: self.terms
        }

        # operation
        with pytest.raises(
            AEAEnforceError,
            match="Initial acceptance is already in the list of pending initial acceptances.",
        ):
            self.transactions.add_pending_initial_acceptance(
                self.dialogue_label, self.proposal_id, self.terms,
            )

    def test_add_pending_initial_acceptance_iii(self):
        """Test the add_pending_initial_acceptance method of the Transactions class where dialogue_label and proposal_id IS in _pending_initial_acceptances."""
        # setup
        self.transactions._pending_initial_acceptances[self.dialogue_label] = {
            self.proposal_id: self.terms
        }

        # operation
        with pytest.raises(
            AEAEnforceError,
            match="Initial acceptance is already in the list of pending initial acceptances.",
        ):
            self.transactions.add_pending_initial_acceptance(
                self.dialogue_label, self.proposal_id, self.terms,
            )

    def test_pop_pending_initial_acceptance_i(self):
        """Test the pop_pending_initial_acceptance method of the Transactions class."""
        # setup
        self.transactions.add_pending_initial_acceptance(
            self.dialogue_label, self.proposal_id, self.terms,
        )

        # operation
        actual_terms = self.transactions.pop_pending_initial_acceptance(
            self.dialogue_label, self.proposal_id
        )

        # after
        assert actual_terms == self.terms
        assert (
            self.proposal_id
            not in self.transactions._pending_proposals[self.dialogue_label]
        )

    def test_pop_pending_initial_acceptance_ii(self):
        """Test the pop_pending_initial_acceptance method of the Transactions class where dialogue_label IS in _pending_initial_acceptances."""
        # setup
        self.transactions.add_pending_initial_acceptance(
            self.dialogue_label, self.proposal_id, self.terms,
        )
        self.transactions._pending_initial_acceptances = {}

        # operation
        with pytest.raises(
            AEAEnforceError,
            match="Cannot find the initial acceptance in the list of pending initial acceptances.",
        ):
            assert self.transactions.pop_pending_initial_acceptance(
                self.dialogue_label, self.proposal_id
            )

    def test_pop_pending_initial_acceptance_iii(self):
        """Test the pop_pending_initial_acceptance method of the Transactions class where dialogue_label and proposal_id IS in _pending_initial_acceptances."""
        # setup
        self.transactions.add_pending_initial_acceptance(
            self.dialogue_label, self.proposal_id, self.terms,
        )
        self.transactions._pending_initial_acceptances[self.dialogue_label] = {
            1: self.terms
        }

        # operation
        with pytest.raises(
            AEAEnforceError,
            match="Cannot find the initial acceptance in the list of pending initial acceptances.",
        ):
            assert self.transactions.pop_pending_initial_acceptance(
                self.dialogue_label, self.proposal_id
            )

    def test_register_transaction_with_time(self):
        """Test the _register_transaction_with_time method of the Transactions class."""
        # setup
        datetime_mock = Mock(wraps=datetime.datetime)
        mocked_now = datetime.datetime.strptime("01 01 2020  00:03", "%d %m %Y %H:%M")
        datetime_mock.now.return_value = mocked_now

        # operation
        with patch("datetime.datetime", new=datetime_mock):
            self.transactions._register_transaction_with_time(self.transaction_id)

        # after
        assert (mocked_now, self.transaction_id,)[
            1
        ] == self.transactions._last_update_for_transactions[0][1]

    def test_add_locked_tx_seller(self):
        """Test the add_locked_tx method of the Transactions class as Seller."""
        # setup
        datetime_mock = Mock(wraps=datetime.datetime)
        mocked_now = datetime.datetime.strptime("01 01 2020  00:03", "%d %m %Y %H:%M")
        datetime_mock.now.return_value = mocked_now

        # operation
        with patch("datetime.datetime", new=datetime_mock):
            self.transactions.add_locked_tx(self.terms, FipaDialogue.Role.SELLER)

        # after
        assert (mocked_now, self.terms.id,)[
            1
        ] == self.transactions._last_update_for_transactions[0][1]
        assert self.transactions._locked_txs[self.terms.id] == self.terms
        assert self.transactions._locked_txs_as_seller[self.terms.id] == self.terms
        assert self.terms.id not in self.transactions._locked_txs_as_buyer

    def test_add_locked_tx_buyer(self):
        """Test the add_locked_tx method of the Transactions class as Seller."""
        # setup
        datetime_mock = Mock(wraps=datetime.datetime)
        mocked_now = datetime.datetime.strptime("01 01 2020  00:03", "%d %m %Y %H:%M")
        datetime_mock.now.return_value = mocked_now

        # operation
        with patch("datetime.datetime", new=datetime_mock):
            self.transactions.add_locked_tx(self.terms, FipaDialogue.Role.BUYER)

        # after
        assert (mocked_now, self.terms.id,)[
            1
        ] == self.transactions._last_update_for_transactions[0][1]
        assert self.transactions._locked_txs[self.terms.id] == self.terms
        assert self.transactions._locked_txs_as_buyer[self.terms.id] == self.terms
        assert self.terms.id not in self.transactions._locked_txs_as_seller

    def test_add_locked_tx_fails(self):
        """Test the add_locked_tx method of the Transactions class where transaction_id IS in _locked_txs."""
        # setup
        self.transactions._locked_txs[self.terms.id] = self.terms

        datetime_mock = Mock(wraps=datetime.datetime)
        mocked_now = datetime.datetime.strptime("01 01 2020  00:03", "%d %m %Y %H:%M")
        datetime_mock.now.return_value = mocked_now

        # operation
        with patch("datetime.datetime", new=datetime_mock):
            with pytest.raises(
                AEAEnforceError,
                match="This transaction is already a locked transaction.",
            ):
                self.transactions.add_locked_tx(self.terms, FipaDialogue.Role.BUYER)

        # after
        assert (
            mocked_now,
            self.terms.id,
        ) not in self.transactions._last_update_for_transactions
        assert self.terms.id not in self.transactions._locked_txs_as_buyer
        assert self.terms.id not in self.transactions._locked_txs_as_seller

    def test_pop_locked_tx(self):
        """Test the pop_locked_tx method of the Transactions class."""
        # setup
        self.transactions.add_locked_tx(self.terms, FipaDialogue.Role.BUYER)

        # before
        assert self.terms.id in self.transactions._locked_txs
        assert self.terms.id in self.transactions._locked_txs_as_buyer
        assert self.terms.id not in self.transactions._locked_txs_as_seller

        # operation
        actual_terms = self.transactions.pop_locked_tx(self.terms)

        # after
        assert actual_terms == self.terms

        assert self.terms.id not in self.transactions._locked_txs
        assert self.terms.id not in self.transactions._locked_txs_as_buyer
        assert self.terms.id not in self.transactions._locked_txs_as_seller

    def test_pop_locked_tx_fails(self):
        """Test the pop_locked_tx method of the Transactions class where terms.id is NOT in _locked_txs."""
        # before
        assert self.terms.id not in self.transactions._locked_txs
        assert self.terms.id not in self.transactions._locked_txs_as_buyer
        assert self.terms.id not in self.transactions._locked_txs_as_seller

        # operation
        with pytest.raises(
            AEAEnforceError,
            match="Cannot find this transaction in the list of locked transactions.",
        ):
            self.transactions.pop_locked_tx(self.terms)

        # after
        assert self.terms.id not in self.transactions._locked_txs
        assert self.terms.id not in self.transactions._locked_txs_as_buyer
        assert self.terms.id not in self.transactions._locked_txs_as_seller

    def test_ownership_state_after_locks(self):
        """Test the ownership_state_after_locks method of the Transactions class."""
        # setup
        is_seller = True
        self.transactions._locked_txs_as_seller[self.transaction_id] = self.terms
        expected_apply_transactions_argument = [self.terms]
        expected_ownership_state = OwnershipState()

        # operation
        with patch.object(
            self.skill.skill_context.decision_maker_handler_context.ownership_state,
            "apply_transactions",
            return_value=expected_ownership_state,
        ) as mock_apply_transactions:
            actual_ownership_states = self.transactions.ownership_state_after_locks(
                is_seller
            )

        # after
        mock_apply_transactions.assert_any_call(expected_apply_transactions_argument)
        assert actual_ownership_states == expected_ownership_state



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains the behaviour to get the Fetch random beacon."""

import json
from typing import Any, Dict, Optional, cast

from aea.skills.behaviours import TickerBehaviour

from packages.fetchai.connections.http_client.connection import PUBLIC_ID
from packages.fetchai.protocols.http.message import HttpMessage
from packages.fetchai.skills.fetch_beacon.dialogues import HttpDialogues


DEFAULT_URL = ""


class FetchBeaconBehaviour(TickerBehaviour):
    """This class provides a simple beacon fetch behaviour."""

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the beacon fetch behaviour."""

        super().__init__(**kwargs)
        self.beacon_url = kwargs.pop("beacon_url", DEFAULT_URL)

    def send_http_request_message(
        self, method: str, url: str, content: Optional[Dict] = None
    ) -> None:
        """
        Send an http request message.

        :param method: the http request method (i.e. 'GET' or 'POST').
        :param url: the url to send the message to.
        :param content: the payload.

        :return: None
        """

        # context
        http_dialogues = cast(HttpDialogues, self.context.http_dialogues)

        # http request message
        request_http_message, _ = http_dialogues.create(
            counterparty=str(PUBLIC_ID),
            performative=HttpMessage.Performative.REQUEST,
            method=method,
            url=url,
            headers="",
            version="",
            body=b"" if content is None else json.dumps(content).encode("utf-8"),
        )

        # send message
        self.context.outbox.put_message(message=request_http_message)

    def setup(self) -> None:
        """
        Implement the setup.

        :return: None
        """
        self.context.logger.info("setting up FetchBeaconBehaviour")

    def act(self) -> None:
        """
        Implement the act.

        :return: None
        """

        self.send_http_request_message("GET", self.beacon_url)
        self.context.logger.info(
            "Fetching random beacon from {}...".format(self.beacon_url)
        )

    def teardown(self) -> None:
        """
        Implement the task teardown.

        :return: None
        """
        self.context.logger.info("tearing down FetchBeaconBehaviour")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Bump the AEA version throughout the code base."""

import argparse
import re
import sys
from pathlib import Path

from packaging.version import Version

from aea.configurations.constants import (
    DEFAULT_AEA_CONFIG_FILE,
    DEFAULT_CONNECTION_CONFIG_FILE,
    DEFAULT_CONTRACT_CONFIG_FILE,
    DEFAULT_PROTOCOL_CONFIG_FILE,
    DEFAULT_SKILL_CONFIG_FILE,
)
from aea.helpers.base import compute_specifier_from_version
from scripts.generate_ipfs_hashes import update_hashes


VERSION_NUMBER_PART_REGEX = r"(0|[1-9]\d*)"
VERSION_REGEX = fr"(any|latest|({VERSION_NUMBER_PART_REGEX})\.({VERSION_NUMBER_PART_REGEX})\.({VERSION_NUMBER_PART_REGEX})(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?)"

PACKAGES_DIR = Path("packages")
TESTS_DIR = Path("tests")
AEA_DIR = Path("aea")
CONFIGURATION_FILENAME_REGEX = re.compile(
    "|".join(
        [
            DEFAULT_AEA_CONFIG_FILE,
            DEFAULT_SKILL_CONFIG_FILE,
            DEFAULT_CONNECTION_CONFIG_FILE,
            DEFAULT_CONTRACT_CONFIG_FILE,
            DEFAULT_PROTOCOL_CONFIG_FILE,
        ]
    )
)

IGNORE_DIRS = [Path(".git")]


def update_version_for_files(current_version: str, new_version: str) -> None:
    """
    Update the version.

    :param current_version: the current version
    :param new_version: the new version
    """
    files = [
        Path("benchmark", "run_from_branch.sh"),
        Path("deploy-image", "docker-env.sh"),
        Path("deploy-image", "Dockerfile"),
        Path("develop-image", "docker-env.sh"),
        Path("docs", "quickstart.md"),
        Path("scripts", "install.ps1"),
        Path("scripts", "install.sh"),
        Path("tests", "test_docs", "test_bash_yaml", "md_files", "bash-quickstart.md"),
        Path("user-image", "docker-env.sh"),
    ]
    for filepath in files:
        update_version_for_file(filepath, current_version, new_version)


def update_version_for_aea(new_version: str) -> str:
    """
    Update version for file.

    :param new_version: the new version
    :return: the current version
    """
    current_version = ""
    path = Path("aea", "__version__.py")
    with open(path, "rt") as fin:
        for line in fin:
            if "__version__" not in line:
                continue
            match = re.search(VERSION_REGEX, line)
            if match is None:
                raise ValueError("Current version is not well formatted.")
            current_version = match.group(1)
    if current_version == "":
        raise ValueError("No version found!")
    update_version_for_file(path, current_version, new_version)
    return current_version


def update_version_for_file(path: Path, current_version: str, new_version: str) -> None:
    """
    Update version for file.

    :param path: the file path
    :param current_version: the current version
    :param new_version: the new version
    """
    content = path.read_text()
    content = content.replace(current_version, new_version)
    path.write_text(content)


def update_aea_version_specifiers(old_version: Version, new_version: Version) -> bool:
    """
    Update aea_version specifier set in docs.

    :param old_version: the old version.
    :param new_version: the new version.
    :return: True if the update has been done, False otherwise.
    """
    old_specifier_set = compute_specifier_from_version(old_version)
    new_specifier_set = compute_specifier_from_version(new_version)
    print(f"Old version specifier: {old_specifier_set}")
    print(f"New version specifier: {new_specifier_set}")
    old_specifier_set_regex = re.compile(str(old_specifier_set).replace(" ", " *"))
    if old_specifier_set == new_specifier_set:
        print("Not updating version specifier - they haven't changed.")
        return False
    for file in filter(lambda p: not p.is_dir(), Path(".").rglob("*")):
        dir_root = Path(file.parts[0])
        if dir_root in IGNORE_DIRS:
            print(f"Skipping '{file}'...")
            continue
        print(
            f"Replacing '{old_specifier_set}' with '{new_specifier_set}' in '{file}'... ",
            end="",
        )
        try:
            content = file.read_text()
        except UnicodeDecodeError as e:
            print(f"Cannot read {file}: {str(e)}. Continue...")
        else:
            if old_specifier_set_regex.search(content) is not None:
                content = old_specifier_set_regex.sub(new_specifier_set, content)
                file.write_text(content)
    return True


def parse_args() -> argparse.Namespace:
    """Parse arguments."""

    parser = argparse.ArgumentParser("bump_aea_version")
    parser.add_argument(
        "--new-version", type=str, required=True, help="The new version."
    )
    parser.add_argument("--no-fingerprint", action="store_true")
    arguments_ = parser.parse_args()
    return arguments_


if __name__ == "__main__":
    arguments = parse_args()
    _new_version_str = arguments.new_version

    _current_version_str = update_version_for_aea(_new_version_str)
    update_version_for_files(_current_version_str, _new_version_str)

    _new_version: Version = Version(_new_version_str)
    _current_version: Version = Version(_current_version_str)
    have_updated_specifier_set = update_aea_version_specifiers(
        _current_version, _new_version
    )

    print("OK")
    return_code = 0
    if arguments.no_fingerprint:
        print("Not updating fingerprints, since --no-fingerprint was specified.")
    elif not have_updated_specifier_set:
        print("Not updating fingerprints, since no specifier set has been updated.")
    else:
        print("Updating hashes and fingerprints.")
        return_code = update_hashes()
    sys.exit(return_code)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains helper function to extract code from the .md files."""
import traceback
from functools import partial
from pathlib import Path
from typing import Dict, Optional

import mistune
import pytest


MISTUNE_BLOCK_CODE_ID = "block_code"


def block_code_filter(b: Dict) -> bool:
    """Check Mistune block is a code block."""
    return b["type"] == MISTUNE_BLOCK_CODE_ID


def type_filter(type_: Optional[str], b: Dict) -> bool:
    """
    Check Mistune code block is of a certain type.

    If the field "info" is None, return False.
    If type_ is None, this function always return true.

    :param type_: the expected type of block (optional)
    :param b: the block dicionary.
    :return: True if the block should be accepted, false otherwise.
    """
    if type_ is None:
        return True
    return b["info"].strip() == type_ if b["info"] is not None else False


def extract_code_blocks(filepath, filter_=None):
    """Extract code blocks from .md files."""
    content = Path(filepath).read_text(encoding="utf-8")
    markdown_parser = mistune.create_markdown(renderer=mistune.AstRenderer())
    blocks = markdown_parser(content)
    actual_type_filter = partial(type_filter, filter_)
    code_blocks = list(filter(block_code_filter, blocks))
    bash_code_blocks = filter(actual_type_filter, code_blocks)
    return list(b["text"] for b in bash_code_blocks)


def extract_python_code(filepath):
    """Removes the license part from the scripts"""
    python_str = ""
    with open(filepath, "r") as python_file:
        read_python_file = python_file.readlines()
    for i in range(21, len(read_python_file)):
        python_str += read_python_file[i]

    return python_str


def compile_and_exec(code: str, locals_dict: Dict = None) -> Dict:
    """
    Compile and exec the code.

    :param code: the code to execute.
    :param locals_dict: the dictionary of local variables.
    :return: the dictionary of locals.
    """
    locals_dict = {} if locals_dict is None else locals_dict
    try:
        code_obj = compile(code, "fakemodule", "exec")
        exec(code_obj, locals_dict)  # nosec
    except Exception:
        pytest.fail(
            "The execution of the following code:\n{}\nfailed with error:\n{}".format(
                code, traceback.format_exc()
            )
        )
    return locals_dict


def compare_enum_classes(expected_enum_class, actual_enum_class):
    """Compare enum classes."""
    try:
        # do some pre-processing
        expected_pairs = sorted(map(lambda x: (x.name, x.value), expected_enum_class))
        actual_pairs = sorted(map(lambda x: (x.name, x.value), actual_enum_class))
        assert expected_pairs == actual_pairs, "{} != {}".format(
            expected_pairs, actual_pairs
        )
    except AssertionError:
        pytest.fail(
            "Actual enum {} is different from the actual one {}".format(
                expected_enum_class, actual_enum_class
            )
        )


class BaseTestMarkdownDocs:
    """Base test class for testing Markdown documents."""

    DOC_PATH: Path

    @classmethod
    def setup_class(cls):
        """Set up the test."""
        markdown_parser = mistune.create_markdown(renderer=mistune.AstRenderer())
        cls.doc_path = cls.DOC_PATH
        cls.doc_content = cls.doc_path.read_text()
        cls.blocks = markdown_parser(cls.doc_content)
        cls.code_blocks = list(filter(block_code_filter, cls.blocks))


class BasePythonMarkdownDocs(BaseTestMarkdownDocs):
    """Test Markdown documentation by running Python snippets in sequence."""

    @classmethod
    def setup_class(cls):
        """
        Set up class.

        It sets the initial value of locals and globals.
        """
        super().setup_class()
        cls.locals = {}
        cls.globals = {}

    @classmethod
    def _python_selector(cls, block: Dict) -> bool:
        return block["type"] == MISTUNE_BLOCK_CODE_ID and (
            block["info"].strip() == "python" if block["info"] else False
        )

    def _assert(self, locals_, *mocks):
        """Do assertions after Python code execution."""

    def test_python_blocks(self, *mocks):
        """Run Python code block in sequence."""
        python_blocks = list(filter(self._python_selector, self.blocks))

        globals_, locals_ = self.globals, self.locals
        for python_block in python_blocks:
            python_code = python_block["text"]
            exec(python_code, globals_, locals_)  # nosec
        self._assert(locals_, *mocks)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for Envelope of mail.base.py."""
import unittest.mock

import pytest

import aea
from aea.configurations.base import PublicId
from aea.exceptions import AEAEnforceError
from aea.mail import base_pb2
from aea.mail.base import Envelope, EnvelopeContext, ProtobufEnvelopeSerializer, URI
from aea.multiplexer import InBox, Multiplexer, OutBox
from aea.protocols.base import Message

from packages.fetchai.connections.local.connection import LocalNode
from packages.fetchai.protocols.default.message import DefaultMessage

from tests.common.utils import wait_for_condition
from tests.conftest import _make_dummy_connection, _make_local_connection


def test_uri():
    """Testing the uri initialisation."""
    uri_raw = "http://user:pwd@NetLoc:80/path;param?query=arg#frag"
    uri = URI(uri_raw=uri_raw)
    assert uri_raw == str(uri)
    assert uri.scheme == "http"
    assert uri.netloc == "user:pwd@NetLoc:80"
    assert uri.path == "/path"
    assert uri.params == "param"
    assert uri.query == "query=arg"
    assert uri.fragment == "frag"
    assert uri.host == "netloc"
    assert uri.port == 80
    assert uri.username == "user"
    assert uri.password == "pwd"


def test_uri_eq():
    """Testing the uri __eq__ function."""
    uri_raw = "http://user:pwd@NetLoc:80/path;param?query=arg#frag"
    uri = URI(uri_raw=uri_raw)
    assert uri == uri


def test_envelope_initialisation():
    """Testing the envelope initialisation."""
    agent_address = "Agent0"
    receiver_address = "Agent1"
    msg = DefaultMessage(
        performative=DefaultMessage.Performative.BYTES, content="hello"
    )
    msg.to = receiver_address

    envelope = Envelope(to=receiver_address, sender=agent_address, message=msg,)

    assert envelope, "Cannot generate a new envelope"

    envelope.to = "ChangedAgent"
    envelope.sender = "ChangedSender"
    envelope.message = b"HelloWorld"

    assert envelope.to == "ChangedAgent", "Cannot set to value on Envelope"
    assert envelope.sender == "ChangedSender", "Cannot set sender value on Envelope"
    assert envelope.message == b"HelloWorld", "Cannot set message on Envelope"
    assert envelope.context is None
    assert not envelope.is_sender_public_id
    assert not envelope.is_to_public_id


def test_inbox_empty():
    """Tests if the inbox is empty."""
    multiplexer = Multiplexer([_make_dummy_connection()])
    _inbox = InBox(multiplexer)
    assert _inbox.empty(), "Inbox is not empty"


def test_inbox_nowait():
    """Tests the inbox without waiting."""
    agent_address = "Agent0"
    receiver_address = "Agent1"
    msg = DefaultMessage(
        performative=DefaultMessage.Performative.BYTES, content="hello"
    )
    msg.to = receiver_address
    multiplexer = Multiplexer([_make_dummy_connection()])
    envelope = Envelope(to=receiver_address, sender=agent_address, message=msg,)
    multiplexer.in_queue.put(envelope)
    inbox = InBox(multiplexer)
    assert (
        inbox.get_nowait() == envelope
    ), "Check for a message on the in queue and wait for no time."


def test_inbox_get():
    """Tests for a envelope on the in queue."""
    agent_address = "Agent0"
    receiver_address = "Agent1"
    msg = DefaultMessage(
        performative=DefaultMessage.Performative.BYTES, content="hello"
    )
    msg.to = receiver_address
    multiplexer = Multiplexer([_make_dummy_connection()])
    envelope = Envelope(to=receiver_address, sender=agent_address, message=msg,)
    multiplexer.in_queue.put(envelope)
    inbox = InBox(multiplexer)

    assert (
        inbox.get() == envelope
    ), "Checks if the returned envelope is the same with the queued envelope."


def test_inbox_get_raises_exception_when_empty():
    """Test that getting an envelope from an empty inbox raises an exception."""
    multiplexer = Multiplexer([_make_dummy_connection()])
    inbox = InBox(multiplexer)

    with pytest.raises(aea.mail.base.Empty):
        with unittest.mock.patch.object(multiplexer, "get", return_value=None):
            inbox.get()


def test_inbox_get_nowait_returns_none():
    """Test that getting an envelope from an empty inbox returns None."""
    # TODO get_nowait in this case should raise an exception, like it's done in queue.Queue
    multiplexer = Multiplexer([_make_dummy_connection()])
    inbox = InBox(multiplexer)
    assert inbox.get_nowait() is None


def test_outbox_put():
    """Tests that an envelope is putted into the queue."""
    agent_address = "Agent0"
    receiver_address = "Agent1"
    msg = DefaultMessage(
        dialogue_reference=("", ""),
        message_id=1,
        target=0,
        performative=DefaultMessage.Performative.BYTES,
        content=b"hello",
    )
    msg.to = receiver_address
    msg.sender = agent_address
    dummy_connection = _make_dummy_connection()
    multiplexer = Multiplexer([dummy_connection])
    outbox = OutBox(multiplexer)
    inbox = InBox(multiplexer)
    multiplexer.connect()
    envelope = Envelope(to=receiver_address, sender=agent_address, message=msg,)
    outbox.put(envelope)
    wait_for_condition(
        lambda: inbox.empty(), 15, "Inbox must not be empty after putting an envelope"
    )
    multiplexer.disconnect()


def test_outbox_put_message():
    """Tests that an envelope is created from the message is in the queue."""
    agent_address = "Agent0"
    receiver_address = "Agent1"
    msg = DefaultMessage(
        dialogue_reference=("", ""),
        message_id=1,
        target=0,
        performative=DefaultMessage.Performative.BYTES,
        content=b"hello",
    )
    msg.to = receiver_address
    msg.sender = agent_address
    dummy_connection = _make_dummy_connection()
    multiplexer = Multiplexer([dummy_connection])
    outbox = OutBox(multiplexer)
    inbox = InBox(multiplexer)
    multiplexer.connect()
    wait_for_condition(
        lambda: multiplexer.is_connected, 15, "Multiplexer is not connected"
    )
    outbox.put_message(msg)
    wait_for_condition(
        lambda: inbox.empty(), 15, "Inbox must not be empty after putting a message"
    )
    multiplexer.disconnect()


def test_outbox_empty():
    """Test thet the outbox queue is empty."""
    dummy_connection = _make_dummy_connection()
    multiplexer = Multiplexer([dummy_connection])
    multiplexer.connect()
    outbox = OutBox(multiplexer)
    assert outbox.empty(), "The outbox is not empty"
    multiplexer.disconnect()


def test_multiplexer():
    """Tests if the multiplexer is connected."""
    with LocalNode() as node:
        address_1 = "address_1"
        oef_local_connection = _make_local_connection(address_1, node)
        multiplexer = Multiplexer([oef_local_connection])
        multiplexer.connect()
        assert (
            multiplexer.is_connected
        ), "Mailbox cannot connect to the specific Connection(OEFLocalConnection)"
        multiplexer.disconnect()


def test_envelope_fails_on_message_empty_protocol_specification_id():
    """Check message.protocol_specification_id."""

    class BadMessage(Message):
        protocol_id = "some/some:0.1.0"

    message = BadMessage()

    with pytest.raises(ValueError):
        Envelope(message=message, to="1", sender="1")


def test_protobuf_envelope_serializer():
    """Test Protobuf envelope serializer."""
    serializer = ProtobufEnvelopeSerializer()
    # connection id is None because it is not included in the encoded envelope
    envelope_context = EnvelopeContext(connection_id=None, uri=URI("/uri"))
    expected_envelope = Envelope(
        to="to",
        sender="sender",
        protocol_specification_id=PublicId("author", "name", "0.1.0"),
        message=b"message",
        context=envelope_context,
    )
    encoded_envelope = serializer.encode(expected_envelope)
    actual_envelope = serializer.decode(encoded_envelope)

    assert actual_envelope == expected_envelope


def test_envelope_serialization():
    """Test Envelope.encode and Envelope.decode methods."""
    expected_envelope = Envelope(
        to="to",
        sender="sender",
        protocol_specification_id=PublicId("author", "name", "0.1.0"),
        message=b"message",
    )
    encoded_envelope = expected_envelope.encode()
    actual_envelope = Envelope.decode(encoded_envelope)

    assert actual_envelope == expected_envelope


def test_envelope_message_bytes():
    """Test the property Envelope.message_bytes."""
    message = DefaultMessage(DefaultMessage.Performative.BYTES, content=b"message")
    envelope = Envelope(to="to", sender="sender", message=message,)

    expected_message_bytes = message.encode()
    actual_message_bytes = envelope.message_bytes
    assert expected_message_bytes == actual_message_bytes


def test_envelope_context_connection_id():
    """Test the property EnvelopeContext.connection_id."""
    connection_id = PublicId("author", "skill_name", "0.1.0")
    envelope_context = EnvelopeContext()
    envelope_context.connection_id = connection_id
    assert envelope_context.connection_id == connection_id


def test_envelope_context_is_None_for_c2c():
    """Test the Envelope context is None for c2c."""
    envelope_context = EnvelopeContext(
        connection_id=PublicId.from_str("author/connection_name:0.1.0")
    )
    message = DefaultMessage(DefaultMessage.Performative.BYTES, content=b"message")
    with pytest.raises(
        AEAEnforceError,
        match="EnvelopeContext must be None for component to component messages.",
    ):
        Envelope(
            to="some_author/some_name:0.1.0",
            sender="some_author/some_name:0.1.0",
            protocol_specification_id=PublicId("author", "name", "0.1.0"),
            message=message,
            context=envelope_context,
        )


def test_envelope_to_as_public_id():
    """Test the Envelope to_as_public_id method."""
    env = Envelope(
        to="some_author/some_name:0.1.0",
        sender="some_author/some_name:0.1.0",
        protocol_specification_id=PublicId("author", "name", "0.1.0"),
        message=b"message",
    )
    assert env.to_as_public_id is not None


def test_envelope_constructor():
    """Test Envelope constructor checks."""
    Envelope(
        to="to",
        sender="sender",
        message=DefaultMessage(performative=DefaultMessage.Performative.BYTES),
    )
    Envelope(
        to="to",
        sender="sender",
        message=b"",
        protocol_specification_id=DefaultMessage.protocol_specification_id,
    )

    with pytest.raises(
        AEAEnforceError, match="message should be a type of Message or bytes!"
    ):
        Envelope(to="to", sender="sender", message=123)

    with pytest.raises(
        Exception,
        match=r"Message is bytes object, protocol_specification_id must be provided!",
    ):
        Envelope(to="asd", sender="asdasd", message=b"sdfdf")


def test_envelope_context_repr():
    """Check repr for EnvelopeContext."""
    assert str(EnvelopeContext(1, 2)) == "EnvelopeContext(connection_id=1, uri=2)"


def test_envelope_specification_id_translated():
    """Test protocol id to protocol specification id translation and back."""
    protocol_specification_id = PublicId("author", "specification", "0.1.0")

    envelope = Envelope(
        to="to",
        sender="sender",
        protocol_specification_id=protocol_specification_id,
        message=b"",
    )

    assert envelope.protocol_specification_id == protocol_specification_id

    envelope_bytes = envelope.encode()
    envelope_pb = base_pb2.Envelope()
    envelope_pb.ParseFromString(envelope_bytes)

    new_envelope = Envelope.decode(envelope_bytes)
    assert new_envelope.protocol_specification_id == envelope.protocol_specification_id



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the behaviour classes of the confirmation aw3 skill."""

from pathlib import Path
from typing import cast
from unittest.mock import patch

from aea.helpers.search.models import (
    Attribute,
    Constraint,
    ConstraintType,
    DataModel,
    Query,
)

from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.skills.confirmation_aw3.behaviours import SearchBehaviour
from packages.fetchai.skills.confirmation_aw3.strategy import Strategy

from tests.conftest import ROOT_DIR
from tests.test_packages.test_skills.test_confirmation_aw3.intermediate_class import (
    ConfirmationAW3TestCase,
)


class TestSearchBehaviour(ConfirmationAW3TestCase):
    """Test search behaviour of confirmation aw3."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "confirmation_aw3")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()

        cls.search_behaviour = cast(
            SearchBehaviour, cls._skill.skill_context.behaviours.search
        )
        cls.strategy = cast(Strategy, cls._skill.skill_context.strategy)

    def test_act(self):
        """Test the act method of the transaction behaviour."""
        # setup
        self.strategy.is_searching = True

        mock_query = Query(
            [Constraint("some_attribute", ConstraintType("==", "some_service"))],
            DataModel(
                "some_name",
                [
                    Attribute(
                        "some_attribute", str, False, "Some attribute descriptions."
                    )
                ],
            ),
        )

        # operation
        with patch.object(self.strategy, "update_search_query_params") as mock_update:
            with patch.object(
                self.strategy, "get_location_and_service_query", return_value=mock_query
            ):
                self.search_behaviour.act()

        # after
        self.assert_quantity_in_outbox(1)

        mock_update.assert_called_once()

        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=OefSearchMessage,
            performative=OefSearchMessage.Performative.SEARCH_SERVICES,
            to=self.skill.skill_context.search_service_address,
            sender=str(self.skill.skill_context.skill_id),
            query=mock_query,
        )
        assert has_attributes, error_str



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains a behaviour for fetching a coin price from an API."""

import json
from typing import Any, Dict, cast

from aea.skills.behaviours import TickerBehaviour

from packages.fetchai.connections.http_client.connection import (
    PUBLIC_ID as HTTP_CLIENT_ID,
)
from packages.fetchai.connections.prometheus.connection import (
    PUBLIC_ID as PROM_CONNECTION_ID,
)
from packages.fetchai.protocols.http.message import HttpMessage
from packages.fetchai.protocols.prometheus.message import PrometheusMessage
from packages.fetchai.skills.coin_price.dialogues import (
    HttpDialogues,
    PrometheusDialogues,
)
from packages.fetchai.skills.coin_price.models import CoinPriceModel


class CoinPriceBehaviour(TickerBehaviour):
    """This class provides a simple behaviour to fetch a coin price."""

    def __init__(self, **kwargs: Any):
        """Initialize the coin price behaviour."""

        super().__init__(**kwargs)

    def send_http_request_message(
        self, method: str, url: str, content: Dict = None
    ) -> None:
        """
        Send an http request message.

        :param method: the http request method (i.e. 'GET' or 'POST').
        :param url: the url to send the message to.
        :param content: the payload.

        :return: None
        """

        # context
        http_dialogues = cast(HttpDialogues, self.context.http_dialogues)

        # http request message
        request_http_message, _ = http_dialogues.create(
            counterparty=str(HTTP_CLIENT_ID),
            performative=HttpMessage.Performative.REQUEST,
            method=method,
            url=url,
            headers="",
            version="",
            body=b"" if content is None else json.dumps(content).encode("utf-8"),
        )

        # send message
        self.context.outbox.put_message(message=request_http_message)

    def add_prometheus_metric(
        self,
        metric_name: str,
        metric_type: str,
        description: str,
        labels: Dict[str, str],
    ) -> None:
        """
        Add a prometheus metric.

        :param metric_name: the name of the metric to add.
        :param type: the type of the metric.
        :param description: a description of the metric.
        :param labels: the metric labels.
        :return: None
        """

        # context
        prom_dialogues = cast(PrometheusDialogues, self.context.prometheus_dialogues)

        # prometheus update message
        message, _ = prom_dialogues.create(
            counterparty=str(PROM_CONNECTION_ID),
            performative=PrometheusMessage.Performative.ADD_METRIC,
            type=metric_type,
            title=metric_name,
            description=description,
            labels=labels,
        )

        # send message
        self.context.outbox.put_message(message=message)

    def update_prometheus_metric(
        self, metric_name: str, update_func: str, value: float, labels: Dict[str, str],
    ) -> None:
        """
        Update a prometheus metric.

        :param metric_name: the name of the metric.
        :param update_func: the name of the update function (e.g. inc, dec, set, ...).
        :param value: the value to provide to the update function.
        :param labels: the metric labels.
        :return: None
        """

        # context
        prom_dialogues = cast(PrometheusDialogues, self.context.prometheus_dialogues)

        # prometheus update message
        message, _ = prom_dialogues.create(
            counterparty=str(PROM_CONNECTION_ID),
            performative=PrometheusMessage.Performative.UPDATE_METRIC,
            title=metric_name,
            callable=update_func,
            value=value,
            labels=labels,
        )

        # send message
        self.context.outbox.put_message(message=message)

    def setup(self) -> None:
        """
        Implement the setup.

        :return: None
        """
        self.context.logger.info("setting up CoinPriceBehaviour")

        prom_dialogues = cast(PrometheusDialogues, self.context.prometheus_dialogues)

        if prom_dialogues.enabled:
            for metric in prom_dialogues.metrics:
                metric_name = metric["name"]
                self.context.logger.info("Adding Prometheus metric: " + metric_name)
                self.add_prometheus_metric(
                    metric_name,
                    metric["type"],
                    metric["description"],
                    dict(metric["labels"]),
                )

    def act(self) -> None:
        """
        Implement the act.

        :return: None
        """
        model = cast(CoinPriceModel, self.context.coin_price_model)

        self.context.logger.info(
            f"Fetching price of {model.coin_id} in {model.currency} from {model.url}"
        )

        url = f"{model.url}simple/price?ids={model.coin_id}&vs_currencies={model.currency}"

        self.send_http_request_message("GET", url)

    def teardown(self) -> None:
        """
        Implement the task teardown.

        :return: None
        """
        self.context.logger.info("tearing down CoinPriceBehaviour")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Module for the multiplexer class and related classes."""
import asyncio
import queue
import threading
from asyncio.events import AbstractEventLoop
from concurrent.futures._base import CancelledError
from concurrent.futures._base import TimeoutError as FuturesTimeoutError
from contextlib import suppress
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)

from aea.common import Address
from aea.configurations.base import PublicId
from aea.connections.base import Connection, ConnectionStates
from aea.exceptions import enforce
from aea.helpers.async_friendly_queue import AsyncFriendlyQueue
from aea.helpers.async_utils import AsyncState, Runnable, ThreadedAsyncRunner
from aea.helpers.exception_policy import ExceptionPolicyEnum
from aea.helpers.logging import WithLogger, get_logger
from aea.mail.base import AEAConnectionError, Empty, Envelope, EnvelopeContext
from aea.protocols.base import Message, Protocol


class MultiplexerStatus(AsyncState):
    """The connection status class."""

    def __init__(self) -> None:
        """Initialize the connection status."""
        super().__init__(
            initial_state=ConnectionStates.disconnected, states_enum=ConnectionStates
        )

    @property
    def is_connected(self) -> bool:  # pragma: nocover
        """Return is connected."""
        return self.get() == ConnectionStates.connected

    @property
    def is_connecting(self) -> bool:  # pragma: nocover
        """Return is connecting."""
        return self.get() == ConnectionStates.connecting

    @property
    def is_disconnected(self) -> bool:  # pragma: nocover
        """Return is disconnected."""
        return self.get() == ConnectionStates.disconnected

    @property
    def is_disconnecting(self) -> bool:  # pragma: nocover
        """Return is disconnected."""
        return self.get() == ConnectionStates.disconnecting


class AsyncMultiplexer(Runnable, WithLogger):
    """This class can handle multiple connections at once."""

    DISCONNECT_TIMEOUT = 5

    def __init__(
        self,
        connections: Optional[Sequence[Connection]] = None,
        default_connection_index: int = 0,
        loop: Optional[AbstractEventLoop] = None,
        exception_policy: ExceptionPolicyEnum = ExceptionPolicyEnum.propagate,
        threaded: bool = False,
        agent_name: str = "standalone",
        default_routing: Optional[Dict[PublicId, PublicId]] = None,
        default_connection: Optional[PublicId] = None,
        protocols: Optional[List[Union[Protocol, Message]]] = None,
    ) -> None:
        """
        Initialize the connection multiplexer.

        :param connections: a sequence of connections.
        :param default_connection_index: the index of the connection to use as default.
            This information is used for envelopes which don't specify any routing context.
            If connections is None, this parameter is ignored.
        :param loop: the event loop to run the multiplexer. If None, a new event loop is created.
        :param agent_name: the name of the agent that owns the multiplexer, for logging purposes.
        """
        self._exception_policy: ExceptionPolicyEnum = exception_policy
        logger = get_logger(__name__, agent_name)
        WithLogger.__init__(self, logger=logger)
        Runnable.__init__(self, loop=loop, threaded=threaded)

        self._connections: List[Connection] = []
        self._id_to_connection: Dict[PublicId, Connection] = {}
        self._default_connection: Optional[Connection] = None

        connections = connections or []
        if not default_connection and connections:
            enforce(
                len(connections) - 1 >= default_connection_index,
                "default_connection_index os out of connections range!",
            )
            default_connection = connections[default_connection_index].connection_id

        if default_connection:
            enforce(
                bool(
                    [
                        i.connection_id.same_prefix(default_connection)
                        for i in connections
                    ]
                ),
                f"Default connection {default_connection} does not present in connections list!",
            )

        self._default_routing = {}  # type: Dict[PublicId, PublicId]

        self._setup(connections or [], default_routing, default_connection)

        self._connection_status = MultiplexerStatus()
        self._specification_id_to_protocol_id = {
            p.protocol_specification_id: p.protocol_id for p in protocols or []
        }
        self._routing_helper: Dict[Address, PublicId] = {}

        self._in_queue = AsyncFriendlyQueue()  # type: AsyncFriendlyQueue
        self._out_queue = None  # type: Optional[asyncio.Queue]

        self._recv_loop_task = None  # type: Optional[asyncio.Task]
        self._send_loop_task = None  # type: Optional[asyncio.Task]

        self._loop: asyncio.AbstractEventLoop = loop if loop is not None else asyncio.new_event_loop()
        self._lock: asyncio.Lock = asyncio.Lock(loop=self._loop)
        self.set_loop(self._loop)

    @property
    def default_connection(self) -> Optional[Connection]:
        """Get the default connection."""
        return self._default_connection

    @property
    def in_queue(self) -> AsyncFriendlyQueue:
        """Get the in queue."""
        return self._in_queue

    @property
    def out_queue(self) -> asyncio.Queue:
        """Get the out queue."""
        if self._out_queue is None:  # pragma: nocover
            raise ValueError("Accessing out queue before loop is started.")
        return self._out_queue

    @property
    def connections(self) -> Tuple[Connection, ...]:
        """Get the connections."""
        return tuple(self._connections)

    @property
    def is_connected(self) -> bool:
        """Check whether the multiplexer is processing envelopes."""
        return self.connection_status.is_connected

    @property
    def default_routing(self) -> Dict[PublicId, PublicId]:
        """Get the default routing."""
        return self._default_routing

    @default_routing.setter
    def default_routing(self, default_routing: Dict[PublicId, PublicId]) -> None:
        """Set the default routing."""
        self._default_routing = default_routing

    @property
    def connection_status(self) -> MultiplexerStatus:
        """Get the connection status."""
        return self._connection_status

    async def run(self) -> None:
        """Run multiplexer connect and recv/send tasks."""
        self.set_loop(asyncio.get_event_loop())
        try:
            await self.connect()

            if not self._recv_loop_task or not self._send_loop_task:
                raise ValueError("Multiplexer is not connected properly.")

            await asyncio.gather(self._recv_loop_task, self._send_loop_task)
        finally:
            await self.disconnect()

    def _get_protocol_id_for_envelope(self, envelope: Envelope) -> PublicId:
        """Get protocol id for envelope."""
        if isinstance(envelope.message, Message):
            return cast(Message, envelope.message).protocol_id

        protocol_id = self._specification_id_to_protocol_id.get(
            envelope.protocol_specification_id
        )

        if not protocol_id:
            raise ValueError(
                f"Can not resolve protocol id for {envelope}, pass protocols supported to multipelxer instance {self._specification_id_to_protocol_id}"
            )

        return protocol_id

    def set_loop(self, loop: AbstractEventLoop) -> None:
        """
        Set event loop and all event loopp related objects.

        :param loop: asyncio event loop.
        :return: None
        """
        self._loop = loop
        self._lock = asyncio.Lock(loop=self._loop)

    def _handle_exception(self, fn: Callable, exc: Exception) -> None:
        """
        Handle exception raised.

        :param fn: a method where it raised .send .connect etc
        :param exc:  exception

        :return: None.
        """
        if self._exception_policy == ExceptionPolicyEnum.just_log:
            self.logger.exception(f"Exception raised in {fn}")
        elif self._exception_policy == ExceptionPolicyEnum.propagate:
            raise exc
        elif self._exception_policy == ExceptionPolicyEnum.stop_and_exit:
            self._loop.create_task(AsyncMultiplexer.disconnect(self))
        else:  # pragma: nocover
            raise ValueError(f"Unknown exception policy: {self._exception_policy}")

    def add_connection(self, connection: Connection, is_default: bool = False) -> None:
        """
        Add a connection to the mutliplexer.

        :param connection: the connection to add.
        :param is_default: whether the connection added should be the default one.
        :return: None
        """
        if connection.connection_id in self._id_to_connection:  # pragma: nocover
            self.logger.warning(
                f"A connection with id {connection.connection_id} was already added. Replacing it..."
            )

        self._connections.append(connection)
        self._id_to_connection[connection.connection_id] = connection
        if is_default:
            self._default_connection = connection

    def _connection_consistency_checks(self) -> None:
        """
        Do some consistency checks on the multiplexer connections.

        :return: None
        :raise AEAEnforceError: if an inconsistency is found.
        """
        if len(self.connections) == 0:
            self.logger.debug("List of connections is empty.")

        enforce(
            len(set(c.connection_id for c in self.connections))
            == len(self.connections),
            "Connection names must be unique.",
        )

    def _set_default_connection_if_none(self) -> None:
        """Set the default connection if it is none."""
        if self._default_connection is None and bool(self.connections):
            self._default_connection = self.connections[0]

    async def connect(self) -> None:
        """Connect the multiplexer."""
        self._loop = asyncio.get_event_loop()
        self.logger.debug("Multiplexer connecting...")
        self._connection_consistency_checks()
        self._set_default_connection_if_none()
        self._out_queue = asyncio.Queue()

        async with self._lock:
            if self.connection_status.is_connected:
                self.logger.debug("Multiplexer already connected.")
                return
            try:
                self.connection_status.set(ConnectionStates.connecting)
                await self._connect_all()

                if all(c.is_connected for c in self._connections):
                    self.connection_status.set(ConnectionStates.connected)
                else:  # pragma: nocover
                    raise AEAConnectionError("Failed to connect the multiplexer.")

                self._recv_loop_task = self._loop.create_task(self._receiving_loop())
                self._send_loop_task = self._loop.create_task(self._send_loop())
                self.logger.debug("Multiplexer connected and running.")
            except (CancelledError, asyncio.CancelledError):  # pragma: nocover
                await self._stop()
                raise asyncio.CancelledError()
            except Exception:
                self.logger.exception("Exception on connect:")
                await self._stop()
                raise AEAConnectionError("Failed to connect the multiplexer.")

    async def disconnect(self) -> None:
        """Disconnect the multiplexer."""
        self.logger.debug("Multiplexer disconnecting...")
        async with self._lock:
            if self.connection_status.is_disconnected:
                self.logger.debug("Multiplexer already disconnected.")
                return
            try:
                self.connection_status.set(ConnectionStates.disconnecting)
                await asyncio.wait_for(self._stop(), timeout=60)
                self.logger.debug("Multiplexer disconnected.")
            except (CancelledError, Exception):
                self.logger.exception("Exception on disconnect:")
                raise AEAConnectionError("Failed to disconnect the multiplexer.")

    async def _stop_receive_send_loops(self) -> None:
        """Stop receive and send loops."""
        self.logger.debug("Stopping recv loop...")

        if self._recv_loop_task:
            self._recv_loop_task.cancel()
            with suppress(Exception, asyncio.CancelledError):
                await self._recv_loop_task

        self._recv_loop_task = None
        self.logger.debug("Recv loop stopped.")

        self.logger.debug("Stopping send loop...")

        if self._send_loop_task:
            # send a 'stop' token (a None value) to wake up the coroutine waiting for outgoing envelopes.
            await self.out_queue.put(None)
            self._send_loop_task.cancel()
            with suppress(Exception, asyncio.CancelledError):
                await self._send_loop_task

        self._send_loop_task = None
        self.logger.debug("Send loop stopped.")

    def _check_and_set_disconnected_state(self) -> None:
        """Check every connection is disconnected and set disconnected state."""
        if all([c.is_disconnected for c in self.connections]):
            self.connection_status.set(ConnectionStates.disconnected)
        else:
            raise AEAConnectionError(
                "Failed to disconnect multiplexer, some connections are not disconnected!"
            )

    async def _stop(self) -> None:
        """
        Stop the multiplexer.

        Stops recv and send loops.
        Disconnect every connection.
        """
        self.logger.debug("Stopping multiplexer...")

        await asyncio.wait_for(self._stop_receive_send_loops(), timeout=60)
        await asyncio.wait_for(self._disconnect_all(), timeout=60)
        self._check_and_set_disconnected_state()

        self.logger.debug("Multiplexer stopped.")

    async def _connect_all(self) -> None:
        """Set all the connection up."""
        self.logger.debug("Starting multiplexer connections.")
        connected = []  # type: List[PublicId]
        for connection_id, connection in self._id_to_connection.items():
            try:
                await self._connect_one(connection_id)
                connected.append(connection_id)
            except Exception as e:  # pylint: disable=broad-except
                if not isinstance(e, (asyncio.CancelledError, CancelledError)):
                    self.logger.exception(
                        "Error while connecting {}: {}".format(
                            str(type(connection)), repr(e)
                        )
                    )
                raise
        self.logger.debug("Multiplexer connections are set.")

    async def _connect_one(self, connection_id: PublicId) -> None:
        """
        Set a connection up.

        :param connection_id: the id of the connection.
        :return: None
        """
        connection = self._id_to_connection[connection_id]
        self.logger.debug("Processing connection {}".format(connection.connection_id))
        if connection.is_connected:
            self.logger.debug(
                "Connection {} already established.".format(connection.connection_id)
            )
        else:
            await connection.connect()
            self.logger.debug(
                "Connection {} has been set up successfully.".format(
                    connection.connection_id
                )
            )

    async def _disconnect_all(self) -> None:
        """Tear all the connections down."""
        self.logger.debug("Tear the multiplexer connections down.")
        for connection_id, connection in self._id_to_connection.items():
            try:
                await asyncio.wait_for(
                    self._disconnect_one(connection_id), timeout=self.DISCONNECT_TIMEOUT
                )
            except FuturesTimeoutError:
                self.logger.debug(  # pragma: nocover
                    f"Disconnection of `{connection_id}` timed out."
                )
            except Exception as e:  # pylint: disable=broad-except
                self.logger.exception(
                    "Error while disconnecting {}: {}".format(
                        str(type(connection)), str(e)
                    )
                )

    async def _disconnect_one(self, connection_id: PublicId) -> None:
        """
        Tear a connection down.

        :param connection_id: the id of the connection.
        :return: None
        """
        connection = self._id_to_connection[connection_id]
        self.logger.debug("Processing connection {}".format(connection.connection_id))
        if not connection.is_connected:
            self.logger.debug(
                "Connection {} already disconnected.".format(connection.connection_id)
            )
        else:
            await connection.disconnect()
            self.logger.debug(
                "Connection {} has been disconnected successfully.".format(
                    connection.connection_id
                )
            )

    async def _send_loop(self) -> None:
        """Process the outgoing envelopes."""
        if not self.is_connected:
            self.logger.debug(
                "Sending loop not started. The multiplexer is not connected."
            )
            return

        try:
            while self.is_connected:
                self.logger.debug("Waiting for outgoing envelopes...")
                envelope = await self.out_queue.get()
                if envelope is None:  # pragma: nocover
                    self.logger.debug(
                        "Received empty envelope. Quitting the sending loop..."
                    )
                    return None
                self.logger.debug("Sending envelope {}".format(str(envelope)))
                await self._send(envelope)

        except asyncio.CancelledError:
            self.logger.debug("Sending loop cancelled.")
            raise
        except Exception as e:  # pylint: disable=broad-except  # pragma: nocover
            self.logger.exception("Error in the sending loop: {}".format(str(e)))
            raise

    async def _receiving_loop(self) -> None:
        """Process incoming envelopes."""
        self.logger.debug("Starting receving loop...")
        task_to_connection = {
            asyncio.ensure_future(conn.receive()): conn for conn in self.connections
        }

        try:
            while self.connection_status.is_connected and len(task_to_connection) > 0:
                done, _pending = await asyncio.wait(
                    task_to_connection.keys(), return_when=asyncio.FIRST_COMPLETED
                )

                # process completed receiving tasks.
                for task in done:
                    connection = task_to_connection.pop(task)
                    envelope = task.result()
                    if envelope is not None:
                        self._update_routing_helper(envelope, connection)
                        self.in_queue.put_nowait(envelope)

                    # reinstantiate receiving task, but only if the connection is still up.
                    if connection.is_connected:
                        new_task = asyncio.ensure_future(connection.receive())
                        task_to_connection[new_task] = connection

        except asyncio.CancelledError:  # pragma: nocover
            self.logger.debug("Receiving loop cancelled.")
            raise
        except Exception as e:  # pylint: disable=broad-except
            self.logger.exception("Error in the receiving loop: {}".format(str(e)))
            raise
        finally:
            # cancel all the receiving tasks.
            for t in task_to_connection.keys():
                t.cancel()
            self.logger.debug("Receiving loop terminated.")

    async def _send(self, envelope: Envelope) -> None:
        """
        Send an envelope.

        :param envelope: the envelope to send.
        :return: None
        :raises ValueError: if the connection id provided is not valid.
        """
        envelope_protocol_id = self._get_protocol_id_for_envelope(envelope)
        connection_id = self._get_connection_id_from_envelope(
            envelope, envelope_protocol_id
        )

        connection = (
            self._get_connection(connection_id) if connection_id is not None else None
        )

        if connection is None:
            # we don't raise on dropping envelope as this can be a configuration issue only!
            self.logger.warning(
                f"Dropping envelope, no connection available for sending: {envelope}"
            )
            return

        if not self._is_connection_supported_protocol(connection, envelope_protocol_id):
            return

        try:
            await connection.send(envelope)
        except Exception as e:  # pylint: disable=broad-except
            self._handle_exception(self._send, e)

    def _get_connection_id_from_envelope(
        self, envelope: Envelope, envelope_protocol_id: PublicId
    ) -> Optional[PublicId]:
        """
        Get the connection id from an envelope.

        Applies the following rules:
        - component to component messages are routed by their component id
        - agent to agent messages, are routed following four rules:
            * first, try to route by envelope context connection id
            * second, try to route by routing helper
            * third, try to route by default routing
            * forth, using default connection

        :param envelope: the Envelope
        :param envelope_protocol_id: the protocol id of the message contained in the envelope
        :return: public id if found
        """
        # component to component messages are routed by their component id
        if envelope.is_component_to_component_message:
            connection_id = envelope.to_as_public_id
            self.logger.debug(
                "Using envelope `to` field as connection_id: {}".format(connection_id)
            )
            enforce(
                connection_id is not None,
                "Connection id cannot be None by envelope construction.",
            )
            return connection_id

        # agent to agent messages, are routed following four rules:
        # first, try to route by envelope context connection id
        if envelope.context is not None and envelope.context.connection_id is not None:
            connection_id = envelope.context.connection_id
            self.logger.debug(
                "Using envelope context connection_id: {}".format(connection_id)
            )
            return connection_id

        # second, try to route by routing helper
        if envelope.to in self._routing_helper:
            connection_id = self._routing_helper[envelope.to]
            self.logger.debug(
                "Using routing helper with connection_id: {}".format(connection_id)
            )
            return connection_id

        # third, try to route by default routing
        if envelope_protocol_id in self.default_routing:
            connection_id = self.default_routing[envelope_protocol_id]
            self.logger.debug("Using default routing: {}".format(connection_id))
            return connection_id

        # forth, using default connection
        connection_id = (
            self.default_connection.connection_id
            if self.default_connection is not None
            else None
        )
        self.logger.debug("Using default connection: {}".format(connection_id))
        return connection_id

    def _get_connection(self, connection_id: PublicId) -> Optional[Connection]:
        """Check if the connection id is registered."""
        conn_ = self._id_to_connection.get(connection_id, None)
        if conn_ is not None:
            return conn_
        for id_, conn_ in self._id_to_connection.items():
            if id_.same_prefix(connection_id):
                return conn_
        self.logger.error(f"No connection registered with id: {connection_id}")
        return None

    def _is_connection_supported_protocol(
        self, connection: Connection, protocol_id: PublicId
    ) -> bool:
        """Check protocol id is supported by the connection."""
        if protocol_id in connection.excluded_protocols:
            self.logger.warning(
                f"Connection {connection.connection_id} does not support protocol {protocol_id}. It is explicitly excluded."
            )
            return False

        if (
            connection.restricted_to_protocols
            and protocol_id not in connection.restricted_to_protocols
        ):
            self.logger.warning(
                f"Connection {connection.connection_id} does not support protocol {protocol_id}. The connection is restricted to protocols in {connection.restricted_to_protocols}."
            )
            return False

        return True

    def get(
        self, block: bool = False, timeout: Optional[float] = None
    ) -> Optional[Envelope]:
        """
        Get an envelope within a timeout.

        :param block: make the call blocking (ignore the timeout).
        :param timeout: the timeout to wait until an envelope is received.
        :return: the envelope, or None if no envelope is available within a timeout.
        """
        try:
            return self.in_queue.get(block=block, timeout=timeout)
        except queue.Empty:
            raise Empty

    async def async_get(self) -> Envelope:
        """
        Get an envelope async way.

        :return: the envelope
        """
        try:
            return await self.in_queue.async_get()
        except queue.Empty:  # pragma: nocover
            raise Empty

    async def async_wait(self) -> None:
        """
        Get an envelope async way.

        :return: the envelope
        """
        return await self.in_queue.async_wait()

    async def _put(self, envelope: Envelope) -> None:
        """
        Schedule an envelope for sending it.

        Notice that the output queue is an asyncio.Queue which uses an event loop
        running on a different thread than the one used in this function.

        :param envelope: the envelope to be sent.
        :return: None
        """
        await self.out_queue.put(envelope)

    def put(self, envelope: Envelope) -> None:
        """
        Schedule an envelope for sending it.

        Notice that the output queue is an asyncio.Queue which uses an event loop
        running on a different thread than the one used in this function.

        :param envelope: the envelope to be sent.
        :return: None
        """
        if self._threaded:
            self._loop.call_soon_threadsafe(self.out_queue.put_nowait, envelope)
        else:
            self.out_queue.put_nowait(envelope)

    def _setup(
        self,
        connections: Collection[Connection],
        default_routing: Optional[Dict[PublicId, PublicId]] = None,
        default_connection: Optional[PublicId] = None,
    ) -> None:
        """
        Set up the multiplexer.

        :param connections: the connections to use. It will replace the other ones.
        :param default_routing: the default routing.
        :param default_connection: the default connection.
        :return: None.
        """
        self.default_routing = default_routing or {}

        # replace connections
        self._connections = []
        self._id_to_connection = {}

        for c in connections:
            self.add_connection(c, c.public_id == default_connection)

    def _update_routing_helper(
        self, envelope: Envelope, connection: Connection
    ) -> None:
        """
        Update the routing helper.

        Saves the source (connection) of an agent-to-agent envelope.

        :param envelope: the envelope to be updated
        :param connection: the connection
        """
        if envelope.is_component_to_component_message:
            return
        self._routing_helper[envelope.sender] = connection.public_id


class Multiplexer(AsyncMultiplexer):
    """Transit sync multiplexer for compatibility."""

    _thread_was_started: bool
    _is_connected: bool

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Initialize the connection multiplexer.

        :param connections: a sequence of connections.
        :param default_connection_index: the index of the connection to use as default.
                                       | this information is used for envelopes which
                                       | don't specify any routing context.
        :param loop: the event loop to run the multiplexer. If None, a new event loop is created.
        """
        super().__init__(*args, **kwargs)
        self._sync_lock = threading.Lock()
        self._init()

    def _init(self) -> None:
        """Set initial variables."""
        self._thread_was_started = False
        self._is_connected = False

    def set_loop(self, loop: AbstractEventLoop) -> None:
        """
        Set event loop and all event loopp related objects.

        :param loop: asyncio event loop.
        :return: None
        """
        super().set_loop(loop)
        self._thread_runner = ThreadedAsyncRunner(self._loop)

    def connect(self) -> None:  # type: ignore # cause overrides coroutine # pylint: disable=invalid-overridden-method
        """
        Connect the multiplexer.

        Synchronously in thread spawned if new loop created.
        """
        with self._sync_lock:
            if not self._loop.is_running():
                self._thread_runner.start()
                self._thread_was_started = True

            self._thread_runner.call(super().connect()).result(240)
            self._is_connected = True

    def disconnect(self) -> None:  # type: ignore # cause overrides coroutine # pylint: disable=invalid-overridden-method
        """
        Disconnect the multiplexer.

        Also stops a dedicated thread for event loop if spawned on connect.
        """
        self.logger.debug("Disconnect called")
        with self._sync_lock:
            if not self._loop.is_running():
                return

            if self._is_connected:
                self._thread_runner.call(super().disconnect()).result(240)
                self._is_connected = False
            self.logger.debug("Disconnect async method executed")

            if self._thread_runner.is_alive() and self._thread_was_started:
                self._thread_runner.stop()
                self.logger.debug("Thread stopped")

            self.logger.debug("Disconnected")

            # reset thread runner and init variables
            self._init()
            self.set_loop(self._loop)

    def put(self, envelope: Envelope) -> None:  # type: ignore  # cause overrides coroutine
        """
        Schedule an envelope for sending it.

        Notice that the output queue is an asyncio.Queue which uses an event loop
        running on a different thread than the one used in this function.

        :param envelope: the envelope to be sent.
        :return: None
        """
        self._thread_runner.call(super()._put(envelope))  # .result(240)


class InBox:
    """A queue from where you can only consume envelopes."""

    def __init__(self, multiplexer: AsyncMultiplexer) -> None:
        """
        Initialize the inbox.

        :param multiplexer: the multiplexer
        """
        super().__init__()
        self._multiplexer = multiplexer

    def empty(self) -> bool:
        """
        Check for a envelope on the in queue.

        :return: boolean indicating whether there is an envelope or not
        """
        return self._multiplexer.in_queue.empty()

    def get(self, block: bool = False, timeout: Optional[float] = None) -> Envelope:
        """
        Check for a envelope on the in queue.

        :param block: make the call blocking (ignore the timeout).
        :param timeout: times out the block after timeout seconds.

        :return: the envelope object.
        :raises Empty: if the attempt to get an envelope fails.
        """
        self._multiplexer.logger.debug("Checks for envelope from the in queue...")
        envelope = self._multiplexer.get(block=block, timeout=timeout)

        if envelope is None:  # pragma: nocover
            raise Empty()

        self._multiplexer.logger.debug(f"Incoming {envelope}")
        return envelope

    def get_nowait(self) -> Optional[Envelope]:
        """
        Check for a envelope on the in queue and wait for no time.

        :return: the envelope object
        """
        try:
            return self.get()
        except Empty:  # pragma: nocover
            return None

    async def async_get(self) -> Envelope:
        """
        Check for a envelope on the in queue.

        :return: the envelope object.
        """
        self._multiplexer.logger.debug(
            "Checks for envelope from the in queue async way..."
        )
        envelope = await self._multiplexer.async_get()

        if envelope is None:  # pragma: nocover
            raise Empty()

        self._multiplexer.logger.debug(f"Incoming envelope: {envelope}")
        return envelope

    async def async_wait(self) -> None:
        """
        Check for a envelope on the in queue.

        :return: the envelope object.
        """
        self._multiplexer.logger.debug(
            "Checks for envelope presents in queue async way..."
        )
        await self._multiplexer.async_wait()


class OutBox:
    """A queue from where you can only enqueue envelopes."""

    def __init__(self, multiplexer: AsyncMultiplexer) -> None:
        """
        Initialize the outbox.

        :param multiplexer: the multiplexer
        """
        super().__init__()
        self._multiplexer = multiplexer

    def empty(self) -> bool:
        """
        Check for a envelope on the in queue.

        :return: boolean indicating whether there is an envelope or not
        """
        return self._multiplexer.out_queue.empty()  # pragma: nocover

    def put(self, envelope: Envelope) -> None:
        """
        Put an envelope into the queue.

        :param envelope: the envelope.
        :return: None
        """
        self._multiplexer.logger.debug(f"Put an envelope in the queue: {envelope}.")
        if not isinstance(envelope.message, Message):
            raise ValueError(
                "Only Message type allowed in envelope message field when putting into outbox."
            )
        message = cast(Message, envelope.message)
        if not message.has_to:  # pragma: nocover
            raise ValueError("Provided message has message.to not set.")
        if not message.has_sender:  # pragma: nocover
            raise ValueError("Provided message has message.sender not set.")
        self._multiplexer.put(envelope)

    def put_message(
        self, message: Message, context: Optional[EnvelopeContext] = None,
    ) -> None:
        """
        Put a message in the outbox.

        This constructs an envelope with the input arguments.

        :param message: the message
        :param context: the envelope context
        :return: None
        """
        if not isinstance(message, Message):
            raise ValueError("Provided message not of type Message.")
        if not message.has_to:
            raise ValueError("Provided message has message.to not set.")
        if not message.has_sender:
            raise ValueError("Provided message has message.sender not set.")
        envelope = Envelope(
            to=message.to, sender=message.sender, message=message, context=context,
        )
        self.put(envelope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains resilience and fault tolerance tests for P2PLibp2p connection."""
import os
import shutil
import tempfile
import time

import pytest

from aea.configurations.constants import DEFAULT_LEDGER
from aea.crypto.registries import make_crypto
from aea.mail.base import Envelope
from aea.multiplexer import Multiplexer

from packages.fetchai.connections.p2p_libp2p.check_dependencies import build_node
from packages.fetchai.protocols.default.message import DefaultMessage
from packages.fetchai.protocols.default.serialization import DefaultSerializer

from tests.common.utils import wait_for_condition
from tests.conftest import (
    MAX_FLAKY_RERUNS_INTEGRATION,
    _make_libp2p_connection,
    libp2p_log_on_failure,
    libp2p_log_on_failure_all,
)


DEFAULT_PORT = 10234


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS_INTEGRATION)
class BaseTestLibp2pRelay:
    """Base test class for libp2p connection relay."""

    @libp2p_log_on_failure
    def setup(self):
        """Set the test up"""
        self.cwd = os.getcwd()
        self.t = tempfile.mkdtemp()
        os.chdir(self.t)
        build_node(self.t)
        self.log_files = []
        self.multiplexers = []

    def change_state_and_wait(
        self,
        multiplexer: Multiplexer,
        expected_is_connected: bool = False,
        timeout: int = 10,
    ) -> None:
        """
        Change state of a multiplexer (either connect or disconnect) and wait.

        :param multiplexer: the multiplexer to connect/disconnect.
        :param expected_is_connected: whether it should be connected or disconnected.
        :param timeout: the maximum number seconds to wait.
        :return: None
        """
        wait_for_condition(
            lambda: multiplexer.is_connected == expected_is_connected, timeout=timeout
        )

    def teardown(self):
        """Tear down the test"""
        for mux in self.multiplexers:
            mux.disconnect()
        os.chdir(self.cwd)
        try:
            shutil.rmtree(self.t)
        except (OSError, IOError):
            pass


@libp2p_log_on_failure_all
class TestLibp2pConnectionRelayNodeRestartIncomingEnvelopes(BaseTestLibp2pRelay):
    """Test that connection will reliably receive envelopes after its relay node restarted"""

    @libp2p_log_on_failure
    def setup(self):
        """Set the test up"""
        super().setup()
        temp_dir_gen = os.path.join(self.t, "temp_dir_gen")
        os.mkdir(temp_dir_gen)
        self.genesis = _make_libp2p_connection(
            data_dir=temp_dir_gen, port=DEFAULT_PORT + 1, build_directory=self.t
        )

        self.multiplexer_genesis = Multiplexer(
            [self.genesis], protocols=[DefaultMessage]
        )
        self.multiplexer_genesis.connect()
        self.log_files.append(self.genesis.node.log_file)
        self.multiplexers.append(self.multiplexer_genesis)

        genesis_peer = self.genesis.node.multiaddrs[0]

        with open("node_key", "wb") as f:
            make_crypto(DEFAULT_LEDGER).dump(f)
            self.relay_key_path = "node_key"

        temp_dir_rel = os.path.join(self.t, "temp_dir_rel")
        os.mkdir(temp_dir_rel)
        self.relay = _make_libp2p_connection(
            data_dir=temp_dir_rel,
            port=DEFAULT_PORT + 2,
            entry_peers=[genesis_peer],
            node_key_file=self.relay_key_path,
            build_directory=self.t,
        )
        self.multiplexer_relay = Multiplexer([self.relay], protocols=[DefaultMessage])
        self.multiplexer_relay.connect()
        self.log_files.append(self.relay.node.log_file)
        self.multiplexers.append(self.multiplexer_relay)

        relay_peer = self.relay.node.multiaddrs[0]

        temp_dir_1 = os.path.join(self.t, "temp_dir_1")
        os.mkdir(temp_dir_1)
        self.connection = _make_libp2p_connection(
            data_dir=temp_dir_1,
            port=DEFAULT_PORT + 3,
            relay=False,
            entry_peers=[relay_peer],
            build_directory=self.t,
        )
        self.multiplexer = Multiplexer([self.connection], protocols=[DefaultMessage])
        self.multiplexer.connect()
        self.log_files.append(self.connection.node.log_file)
        self.multiplexers.append(self.multiplexer)

        temp_dir_2 = os.path.join(self.t, "temp_dir_2")
        os.mkdir(temp_dir_2)
        self.connection2 = _make_libp2p_connection(
            data_dir=temp_dir_2,
            port=DEFAULT_PORT + 4,
            relay=False,
            entry_peers=[relay_peer],
            build_directory=self.t,
        )
        self.multiplexer2 = Multiplexer([self.connection2], protocols=[DefaultMessage])
        self.multiplexer2.connect()
        self.log_files.append(self.connection2.node.log_file)
        self.multiplexers.append(self.multiplexer2)

    def test_connection_is_established(self):
        """Test connection established."""
        assert self.relay.is_connected is True
        assert self.connection.is_connected is True
        assert self.connection2.is_connected is True

    def test_envelope_routed_from_peer_after_relay_restart(self):
        """Test envelope routed from third peer after relay restart."""
        addr_1 = self.genesis.address
        addr_2 = self.connection.address

        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )
        envelope = Envelope(
            to=addr_2,
            sender=addr_1,
            protocol_specification_id=DefaultMessage.protocol_specification_id,
            message=DefaultSerializer().encode(msg),
        )

        self.multiplexer_genesis.put(envelope)
        delivered_envelope = self.multiplexer.get(block=True, timeout=20)

        assert delivered_envelope is not None
        assert delivered_envelope.to == envelope.to
        assert delivered_envelope.sender == envelope.sender
        assert (
            delivered_envelope.protocol_specification_id
            == envelope.protocol_specification_id
        )
        assert delivered_envelope.message_bytes == envelope.message_bytes

        self.multiplexer_relay.disconnect()
        self.change_state_and_wait(self.multiplexer_relay, expected_is_connected=False)

        # currently, multiplexer cannot be restarted
        self.multiplexer_relay = Multiplexer([self.relay], protocols=[DefaultMessage])
        self.multiplexer_relay.connect()
        self.change_state_and_wait(self.multiplexer_relay, expected_is_connected=True)
        self.multiplexers.append(self.multiplexer_relay)

        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"helloAfterRestart",
        )
        envelope = Envelope(
            to=addr_2,
            sender=addr_1,
            protocol_specification_id=DefaultMessage.protocol_specification_id,
            message=DefaultSerializer().encode(msg),
        )

        self.multiplexer_genesis.put(envelope)

        delivered_envelope = self.multiplexer.get(block=True, timeout=20)

        assert delivered_envelope is not None
        assert delivered_envelope.to == envelope.to
        assert delivered_envelope.sender == envelope.sender
        assert (
            delivered_envelope.protocol_specification_id
            == envelope.protocol_specification_id
        )
        assert delivered_envelope.message_bytes == envelope.message_bytes

    def test_envelope_routed_from_client_after_relay_restart(self):
        """Test envelope routed from third relay client after relay restart."""
        addr_1 = self.connection.address
        addr_2 = self.connection2.address

        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )
        envelope = Envelope(
            to=addr_1,
            sender=addr_2,
            protocol_specification_id=DefaultMessage.protocol_specification_id,
            message=DefaultSerializer().encode(msg),
        )

        self.multiplexer2.put(envelope)
        delivered_envelope = self.multiplexer.get(block=True, timeout=20)

        assert delivered_envelope is not None
        assert delivered_envelope.to == envelope.to
        assert delivered_envelope.sender == envelope.sender
        assert (
            delivered_envelope.protocol_specification_id
            == envelope.protocol_specification_id
        )
        assert delivered_envelope.message_bytes == envelope.message_bytes

        self.multiplexer_relay.disconnect()
        self.change_state_and_wait(self.multiplexer_relay, expected_is_connected=False)

        # currently, multiplexer cannot be restarted
        self.multiplexer_relay = Multiplexer([self.relay], protocols=[DefaultMessage])
        self.multiplexer_relay.connect()
        self.change_state_and_wait(self.multiplexer_relay, expected_is_connected=True)
        self.multiplexers.append(self.multiplexer_relay)

        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"helloAfterRestart",
        )

        envelope = Envelope(
            to=addr_1,
            sender=addr_2,
            protocol_specification_id=DefaultMessage.protocol_specification_id,
            message=DefaultSerializer().encode(msg),
        )

        time.sleep(5)
        self.multiplexer2.put(envelope)
        delivered_envelope = self.multiplexer.get(block=True, timeout=20)

        assert delivered_envelope is not None
        assert delivered_envelope.to == envelope.to
        assert delivered_envelope.sender == envelope.sender
        assert (
            delivered_envelope.protocol_specification_id
            == envelope.protocol_specification_id
        )
        assert delivered_envelope.message_bytes == envelope.message_bytes


@libp2p_log_on_failure_all
class TestLibp2pConnectionRelayNodeRestartOutgoingEnvelopes(BaseTestLibp2pRelay):
    """Test that connection will reliably route envelope to destination in case of relay node restart within timeout"""

    @libp2p_log_on_failure
    def setup(self):
        """Set the test up"""
        super().setup()
        temp_dir_gen = os.path.join(self.t, "temp_dir_gen")
        os.mkdir(temp_dir_gen)
        self.genesis = _make_libp2p_connection(
            data_dir=temp_dir_gen, port=DEFAULT_PORT + 1, build_directory=self.t
        )

        self.multiplexer_genesis = Multiplexer(
            [self.genesis], protocols=[DefaultMessage]
        )
        self.multiplexer_genesis.connect()
        self.log_files.append(self.genesis.node.log_file)
        self.multiplexers.append(self.multiplexer_genesis)

        genesis_peer = self.genesis.node.multiaddrs[0]

        with open("node_key", "wb") as f:
            make_crypto(DEFAULT_LEDGER).dump(f)
            self.relay_key_path = "node_key"

        temp_dir_rel = os.path.join(self.t, "temp_dir_rel")
        os.mkdir(temp_dir_rel)
        self.relay = _make_libp2p_connection(
            data_dir=temp_dir_rel,
            port=DEFAULT_PORT + 2,
            entry_peers=[genesis_peer],
            node_key_file=self.relay_key_path,
            build_directory=self.t,
        )
        self.multiplexer_relay = Multiplexer([self.relay], protocols=[DefaultMessage])
        self.multiplexer_relay.connect()
        self.log_files.append(self.relay.node.log_file)
        self.multiplexers.append(self.multiplexer_relay)

        relay_peer = self.relay.node.multiaddrs[0]

        temp_dir_1 = os.path.join(self.t, "temp_dir_1")
        os.mkdir(temp_dir_1)
        self.connection = _make_libp2p_connection(
            data_dir=temp_dir_1,
            port=DEFAULT_PORT + 3,
            relay=False,
            entry_peers=[relay_peer],
            build_directory=self.t,
        )
        self.multiplexer = Multiplexer([self.connection], protocols=[DefaultMessage])
        self.multiplexer.connect()
        self.log_files.append(self.connection.node.log_file)
        self.multiplexers.append(self.multiplexer)

    def test_connection_is_established(self):
        """Test connection established."""
        assert self.relay.is_connected is True
        assert self.connection.is_connected is True

    def test_envelope_routed_after_relay_restart(self):
        """Test envelope routed after relay restart."""
        addr_1 = self.connection.address
        addr_2 = self.genesis.address

        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )
        envelope = Envelope(
            to=addr_2,
            sender=addr_1,
            protocol_specification_id=DefaultMessage.protocol_specification_id,
            message=DefaultSerializer().encode(msg),
        )

        self.multiplexer.put(envelope)
        delivered_envelope = self.multiplexer_genesis.get(block=True, timeout=20)

        assert delivered_envelope is not None
        assert delivered_envelope.to == envelope.to
        assert delivered_envelope.sender == envelope.sender
        assert (
            delivered_envelope.protocol_specification_id
            == envelope.protocol_specification_id
        )
        assert delivered_envelope.message_bytes == envelope.message_bytes

        self.multiplexer_relay.disconnect()
        self.change_state_and_wait(self.multiplexer_relay, expected_is_connected=False)

        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"helloAfterRestart",
        )
        envelope = Envelope(
            to=addr_2,
            sender=addr_1,
            protocol_specification_id=DefaultMessage.protocol_specification_id,
            message=DefaultSerializer().encode(msg),
        )

        self.multiplexer.put(envelope)
        time.sleep(5)

        # currently, multiplexer cannot be restarted
        self.multiplexer_relay = Multiplexer([self.relay], protocols=[DefaultMessage])
        self.multiplexer_relay.connect()
        self.change_state_and_wait(self.multiplexer_relay, expected_is_connected=True)
        self.multiplexers.append(self.multiplexer_relay)

        delivered_envelope = self.multiplexer_genesis.get(block=True, timeout=20)

        assert delivered_envelope is not None
        assert delivered_envelope.to == envelope.to
        assert delivered_envelope.sender == envelope.sender
        assert (
            delivered_envelope.protocol_specification_id
            == envelope.protocol_specification_id
        )
        assert delivered_envelope.message_bytes == envelope.message_bytes


@libp2p_log_on_failure_all
class TestLibp2pConnectionAgentMobility(BaseTestLibp2pRelay):
    """Test that connection will correctly route envelope to destination that changed its peer"""

    @libp2p_log_on_failure
    def setup(self):
        """Set the test up"""
        super().setup()
        temp_dir_gen = os.path.join(self.t, "temp_dir_gen")
        os.mkdir(temp_dir_gen)
        self.genesis = _make_libp2p_connection(data_dir=temp_dir_gen, port=DEFAULT_PORT)

        self.multiplexer_genesis = Multiplexer(
            [self.genesis], protocols=[DefaultMessage]
        )
        self.log_files.append(self.genesis.node.log_file)
        self.multiplexer_genesis.connect()
        self.multiplexers.append(self.multiplexer_genesis)

        genesis_peer = self.genesis.node.multiaddrs[0]

        temp_dir_1 = os.path.join(self.t, "temp_dir_1")
        os.mkdir(temp_dir_1)
        self.connection1 = _make_libp2p_connection(
            data_dir=temp_dir_1, port=DEFAULT_PORT + 1, entry_peers=[genesis_peer]
        )
        self.multiplexer1 = Multiplexer([self.connection1], protocols=[DefaultMessage])
        self.log_files.append(self.connection1.node.log_file)
        self.multiplexer1.connect()
        self.multiplexers.append(self.multiplexer1)

        self.connection_key = make_crypto(DEFAULT_LEDGER)
        temp_dir_2 = os.path.join(self.t, "temp_dir_2")
        os.mkdir(temp_dir_2)
        self.connection2 = _make_libp2p_connection(
            data_dir=temp_dir_2,
            port=DEFAULT_PORT + 2,
            entry_peers=[genesis_peer],
            agent_key=self.connection_key,
        )
        self.multiplexer2 = Multiplexer([self.connection2], protocols=[DefaultMessage])
        self.log_files.append(self.connection2.node.log_file)
        self.multiplexer2.connect()
        self.multiplexers.append(self.multiplexer2)

    def test_connection_is_established(self):
        """Test connection established."""
        assert self.connection1.is_connected is True
        assert self.connection2.is_connected is True

    def test_envelope_routed_after_peer_changed(self):
        """Test envelope routed after peer changed."""
        addr_1 = self.connection1.address
        addr_2 = self.connection2.address

        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )
        envelope = Envelope(
            to=addr_2,
            sender=addr_1,
            protocol_specification_id=DefaultMessage.protocol_specification_id,
            message=DefaultSerializer().encode(msg),
        )

        self.multiplexer1.put(envelope)
        delivered_envelope = self.multiplexer2.get(block=True, timeout=20)

        assert delivered_envelope is not None
        assert delivered_envelope.to == envelope.to
        assert delivered_envelope.sender == envelope.sender
        assert (
            delivered_envelope.protocol_specification_id
            == envelope.protocol_specification_id
        )
        assert delivered_envelope.message_bytes == envelope.message_bytes

        self.multiplexer2.disconnect()
        self.change_state_and_wait(self.multiplexer2, expected_is_connected=False)

        # currently, multiplexer cannot be restarted
        self.multiplexer2 = Multiplexer([self.connection2], protocols=[DefaultMessage])
        self.multiplexer2.connect()
        self.change_state_and_wait(self.multiplexer2, expected_is_connected=True)
        self.multiplexers.append(self.multiplexer2)

        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"helloAfterChangingPeer",
        )
        envelope = Envelope(
            to=addr_2,
            sender=addr_1,
            protocol_specification_id=msg.protocol_specification_id,
            message=msg.encode(),
        )

        self.multiplexer1.put(envelope)

        delivered_envelope = self.multiplexer2.get(block=True, timeout=20)

        assert delivered_envelope is not None
        assert delivered_envelope.to == envelope.to
        assert delivered_envelope.sender == envelope.sender
        assert (
            delivered_envelope.protocol_specification_id
            == envelope.protocol_specification_id
        )
        assert delivered_envelope.message_bytes == envelope.message_bytes



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021 fetchai
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Serialization module for prometheus protocol."""

from typing import Any, Dict, cast

from aea.mail.base_pb2 import DialogueMessage
from aea.mail.base_pb2 import Message as ProtobufMessage
from aea.protocols.base import Message, Serializer

from packages.fetchai.protocols.prometheus import prometheus_pb2
from packages.fetchai.protocols.prometheus.message import PrometheusMessage


class PrometheusSerializer(Serializer):
    """Serialization for the 'prometheus' protocol."""

    @staticmethod
    def encode(msg: Message) -> bytes:
        """
        Encode a 'Prometheus' message into bytes.

        :param msg: the message object.
        :return: the bytes.
        """
        msg = cast(PrometheusMessage, msg)
        message_pb = ProtobufMessage()
        dialogue_message_pb = DialogueMessage()
        prometheus_msg = prometheus_pb2.PrometheusMessage()

        dialogue_message_pb.message_id = msg.message_id
        dialogue_reference = msg.dialogue_reference
        dialogue_message_pb.dialogue_starter_reference = dialogue_reference[0]
        dialogue_message_pb.dialogue_responder_reference = dialogue_reference[1]
        dialogue_message_pb.target = msg.target

        performative_id = msg.performative
        if performative_id == PrometheusMessage.Performative.ADD_METRIC:
            performative = prometheus_pb2.PrometheusMessage.Add_Metric_Performative()  # type: ignore
            type = msg.type
            performative.type = type
            title = msg.title
            performative.title = title
            description = msg.description
            performative.description = description
            labels = msg.labels
            performative.labels.update(labels)
            prometheus_msg.add_metric.CopyFrom(performative)
        elif performative_id == PrometheusMessage.Performative.UPDATE_METRIC:
            performative = prometheus_pb2.PrometheusMessage.Update_Metric_Performative()  # type: ignore
            title = msg.title
            performative.title = title
            callable = msg.callable
            performative.callable = callable
            value = msg.value
            performative.value = value
            labels = msg.labels
            performative.labels.update(labels)
            prometheus_msg.update_metric.CopyFrom(performative)
        elif performative_id == PrometheusMessage.Performative.RESPONSE:
            performative = prometheus_pb2.PrometheusMessage.Response_Performative()  # type: ignore
            code = msg.code
            performative.code = code
            if msg.is_set("message"):
                performative.message_is_set = True
                message = msg.message
                performative.message = message
            prometheus_msg.response.CopyFrom(performative)
        else:
            raise ValueError("Performative not valid: {}".format(performative_id))

        dialogue_message_pb.content = prometheus_msg.SerializeToString()

        message_pb.dialogue_message.CopyFrom(dialogue_message_pb)
        message_bytes = message_pb.SerializeToString()
        return message_bytes

    @staticmethod
    def decode(obj: bytes) -> Message:
        """
        Decode bytes into a 'Prometheus' message.

        :param obj: the bytes object.
        :return: the 'Prometheus' message.
        """
        message_pb = ProtobufMessage()
        prometheus_pb = prometheus_pb2.PrometheusMessage()
        message_pb.ParseFromString(obj)
        message_id = message_pb.dialogue_message.message_id
        dialogue_reference = (
            message_pb.dialogue_message.dialogue_starter_reference,
            message_pb.dialogue_message.dialogue_responder_reference,
        )
        target = message_pb.dialogue_message.target

        prometheus_pb.ParseFromString(message_pb.dialogue_message.content)
        performative = prometheus_pb.WhichOneof("performative")
        performative_id = PrometheusMessage.Performative(str(performative))
        performative_content = dict()  # type: Dict[str, Any]
        if performative_id == PrometheusMessage.Performative.ADD_METRIC:
            type = prometheus_pb.add_metric.type
            performative_content["type"] = type
            title = prometheus_pb.add_metric.title
            performative_content["title"] = title
            description = prometheus_pb.add_metric.description
            performative_content["description"] = description
            labels = prometheus_pb.add_metric.labels
            labels_dict = dict(labels)
            performative_content["labels"] = labels_dict
        elif performative_id == PrometheusMessage.Performative.UPDATE_METRIC:
            title = prometheus_pb.update_metric.title
            performative_content["title"] = title
            callable = prometheus_pb.update_metric.callable
            performative_content["callable"] = callable
            value = prometheus_pb.update_metric.value
            performative_content["value"] = value
            labels = prometheus_pb.update_metric.labels
            labels_dict = dict(labels)
            performative_content["labels"] = labels_dict
        elif performative_id == PrometheusMessage.Performative.RESPONSE:
            code = prometheus_pb.response.code
            performative_content["code"] = code
            if prometheus_pb.response.message_is_set:
                message = prometheus_pb.response.message
                performative_content["message"] = message
        else:
            raise ValueError("Performative not valid: {}.".format(performative_id))

        return PrometheusMessage(
            message_id=message_id,
            dialogue_reference=dialogue_reference,
            target=target,
            performative=performative,
            **performative_content
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains tests for decision_maker."""

import pytest

from aea.error_handler.scaffold import ErrorHandler


def test_scaffold_send_unsupported_protocol_raises_not_implemented_error():
    """Test 'send_unsupported_protocol' raises not implemented error."""
    with pytest.raises(NotImplementedError):
        ErrorHandler.send_unsupported_protocol(None, None)


def test_scaffold_send_decoding_error_raises_not_implemented_error():
    """Test 'send_decoding_error' raises not implemented error."""
    with pytest.raises(NotImplementedError):
        ErrorHandler.send_decoding_error(None, None, None)


def test_scaffold_send_no_active_handler_raises_not_implemented_error():
    """Test 'send_no_active_handler' raises not implemented error."""
    with pytest.raises(NotImplementedError):
        ErrorHandler.send_no_active_handler(None, None, None)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Implementation of the 'aea remove_key' subcommand."""

import os
from typing import cast

import click

from aea.cli.utils.context import Context
from aea.cli.utils.decorators import check_aea_project
from aea.configurations.constants import DEFAULT_AEA_CONFIG_FILE
from aea.crypto.registries import crypto_registry
from aea.helpers.io import open_file


@click.command()
@click.argument(
    "type_",
    metavar="TYPE",
    type=click.Choice(list(crypto_registry.supported_ids)),
    required=True,
)
@click.option(
    "--connection", is_flag=True, help="For removing a private key for connections."
)
@click.pass_context
@check_aea_project
def remove_key(click_context: click.Context, type_: str, connection: bool) -> None:
    """Remove a private key from the wallet of the agent."""
    _remove_private_key(click_context, type_, connection)


def _remove_private_key(
    click_context: click.core.Context, type_: str, connection: bool = False,
) -> None:
    """
    Remove private key to the wallet.

    :param click_context: click context object.
    :param type_: type.
    :param connection: whether or not it is a private key for a connection

    :return: None
    """
    ctx = cast(Context, click_context.obj)
    _try_remove_key(ctx, type_, connection)


def _try_remove_key(ctx: Context, type_: str, connection: bool = False) -> None:
    private_keys = (
        ctx.agent_config.connection_private_key_paths
        if connection
        else ctx.agent_config.private_key_paths
    )
    existing_keys = private_keys.keys()
    if type_ not in existing_keys:
        raise click.ClickException(
            f"There is no {'connection ' if connection else ''}key registered with id {type_}."
        )
    private_keys.delete(type_)
    ctx.agent_loader.dump(
        ctx.agent_config, open_file(os.path.join(ctx.cwd, DEFAULT_AEA_CONFIG_FILE), "w")
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the strategy class of the confirmation aw1 skill."""

import logging
from pathlib import Path
from typing import cast
from unittest.mock import patch

from aea.crypto.ledger_apis import LedgerApis
from aea.helpers.transaction.base import Terms
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.skills.confirmation_aw1.registration_db import RegistrationDB
from packages.fetchai.skills.confirmation_aw1.strategy import (
    PUBLIC_ID,
    REQUIRED_KEYS,
    Strategy,
)

from tests.conftest import ROOT_DIR


class TestStrategy(BaseSkillTestCase):
    """Test Strategy of confirmation aw1."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "confirmation_aw1")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.token_denomination = "atestfet"
        cls.token_dispense_amount = 100000
        cls.fetchai_staking_contract_address = (
            "0x351bac612b50e87b46e4b10a282f632d41397de2"
        )
        cls.override_staking_check = False
        cls.awx_aeas = []
        cls.strategy = Strategy(
            token_denomination=cls.token_denomination,
            token_dispense_amount=cls.token_dispense_amount,
            fetchai_staking_contract_address=cls.fetchai_staking_contract_address,
            override_staking_check=cls.override_staking_check,
            awx_aeas=cls.awx_aeas,
            name="strategy",
            skill_context=cls._skill.skill_context,
        )

        cls.address = "some_address"
        cls.info = {
            "ethereum_address": "some_value",
            "signature_of_ethereum_address": "some_signature_of_ethereum_address",
            "signature_of_fetchai_address": "some_signature_of_fetchai_address",
            "developer_handle": "some_developer_handle",
            "tweet": "some_tweet",
        }
        cls.logger = cls._skill.skill_context.logger
        cls.db = cast(RegistrationDB, cls._skill.skill_context.registration_db)

    def test__init__(self):
        """Test the __init__ of Strategy class."""
        assert self.strategy._is_ready_to_register is False
        assert self.strategy._is_registered is False
        assert self.strategy.is_registration_pending is False
        assert self.strategy.signature_of_ethereum_address is None
        assert self.strategy._ledger_id == self.skill.skill_context.default_ledger_id
        assert self.strategy._max_tx_fee == 100
        assert self.strategy._contract_ledger_id == "ethereum"
        assert self.strategy._contract_callable == "get_stake"
        assert self.strategy._contract_id == str(PUBLIC_ID)
        assert self.strategy._in_process_registrations == {}

    def test_properties(self):
        """Test the properties of Strategy class."""
        assert self.strategy.contract_id == self.strategy._contract_id
        assert self.strategy.contract_address == self.fetchai_staking_contract_address
        assert self.strategy.contract_ledger_id == self.strategy._contract_ledger_id
        assert self.strategy.contract_callable == self.strategy._contract_callable
        assert self.strategy.awx_aeas == self.awx_aeas
        assert self.strategy.all_registered_aeas == []

    def test_lock_registration_temporarily(self):
        """Test the lock_registration_temporarily method of the Strategy class."""
        # before
        assert self.address not in self.strategy._in_process_registrations

        # operation
        self.strategy.lock_registration_temporarily(self.address, self.info)

        # after
        assert self.strategy._in_process_registrations[self.address] == self.info

    def test_finalize_registration(self):
        """Test the finalize_registration method of the Strategy class."""
        # setup
        self.strategy.lock_registration_temporarily(self.address, self.info)

        # operation
        with patch.object(self.db, "set_registered") as mock_set:
            with patch.object(self.logger, "log") as mock_logger:
                self.strategy.finalize_registration(self.address)

        # after
        assert self.address not in self.strategy._in_process_registrations

        mock_logger.assert_any_call(
            logging.INFO,
            f"finalizing registration for address={self.address}, info={self.info}",
        )

        mock_set.assert_any_call(
            address=self.address,
            ethereum_address=self.info["ethereum_address"],
            ethereum_signature=self.info["signature_of_ethereum_address"],
            fetchai_signature=self.info["signature_of_fetchai_address"],
            developer_handle=self.info["developer_handle"],
            tweet=self.info.get("tweet", ""),
        )

    def test_unlock_registration(self):
        """Test the unlock_registration method of the Strategy class."""
        # setup
        self.strategy.lock_registration_temporarily(self.address, self.info)

        # before
        assert self.address in self.strategy._in_process_registrations

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.strategy.unlock_registration(self.address)

        # after
        assert self.address not in self.strategy._in_process_registrations

        mock_logger.assert_any_call(
            logging.INFO,
            f"registration info did not pass staking checks = {self.info}",
        )

    def test_get_developer_handle(self):
        """Test the get_developer_handle method of the Strategy class."""
        # operation
        with patch.object(self.db, "get_developer_handle") as mock_get:
            self.strategy.get_developer_handle(self.address)

        # after
        mock_get.assert_any_call(self.address)

    def test_valid_registration_succeeds(self):
        """Test the valid_registration method of the Strategy class which succeeds."""
        # setup
        registration_info = {
            "ethereum_address": "some_ethereum_address",
            "fetchai_address": self.address,
            "signature_of_ethereum_address": "some_signature_of_ethereum_address",
            "signature_of_fetchai_address": "some_signature_of_fetchai_address",
            "developer_handle": "some_developer_handle",
        }

        # operation
        with patch.object(
            self.strategy, "_valid_signature", return_value=True
        ) as mock_valid:
            with patch.object(self.db, "is_registered", return_value=False) as mock_is:
                is_valid, code, msg = self.strategy.valid_registration(
                    registration_info, self.address
                )

        # after
        mock_is.assert_called_once()
        mock_valid.assert_called()

        assert is_valid
        assert code == 0
        assert msg == "all good!"

    def test_valid_registration_fails_i(self):
        """Test the valid_registration method of the Strategy class which fails because some key is missing."""
        # setup
        incorrect_registration_info = {
            "ethereum_address": "some_ethereum_address",
            "fetchai_address": self.address,
            "signature_of_ethereum_address": "some_signature_of_ethereum_address",
            "signature_of_fetchai_address": "some_signature_of_fetchai_address",
        }

        # operation
        is_valid, code, msg = self.strategy.valid_registration(
            incorrect_registration_info, self.address
        )

        # after
        assert not is_valid
        assert code == 1
        assert msg == f"missing keys in registration info, required: {REQUIRED_KEYS}!"

    def test_valid_registration_fails_ii(self):
        """Test the valid_registration method of the Strategy class which fails because addresses do not match."""
        # setup
        different_addres = "some_other_address"
        incorrect_registration_info = {
            "ethereum_address": "some_ethereum_address",
            "fetchai_address": different_addres,
            "signature_of_ethereum_address": "some_signature_of_ethereum_address",
            "signature_of_fetchai_address": "some_signature_of_fetchai_address",
            "developer_handle": "some_developer_handle",
        }

        # operation
        is_valid, code, msg = self.strategy.valid_registration(
            incorrect_registration_info, self.address
        )

        # after
        assert not is_valid
        assert code == 1
        assert msg == "fetchai address of agent and registration info do not match!"

    def test_valid_registration_fails_iii(self):
        """Test the valid_registration method of the Strategy class which fails because _valid_signature returns False for first call."""
        # setup
        incorrect_registration_info = {
            "ethereum_address": "some_ethereum_address",
            "fetchai_address": self.address,
            "signature_of_ethereum_address": "some_signature_of_ethereum_address",
            "signature_of_fetchai_address": "some_signature_of_fetchai_address",
            "developer_handle": "some_developer_handle",
        }

        # operation
        with patch.object(
            self.strategy, "_valid_signature", return_value=False
        ) as mock_valid:
            is_valid, code, msg = self.strategy.valid_registration(
                incorrect_registration_info, self.address
            )

        # after
        mock_valid.assert_called_once()

        assert not is_valid
        assert code == 1
        assert msg == "fetchai address and signature do not match!"

    def test_valid_registration_fails_iv(self):
        """Test the valid_registration method of the Strategy class which fails because _valid_signature returns False for second call."""
        # setup
        incorrect_registration_info = {
            "ethereum_address": "some_ethereum_address",
            "fetchai_address": self.address,
            "signature_of_ethereum_address": "some_signature_of_ethereum_address",
            "signature_of_fetchai_address": "some_signature_of_fetchai_address",
            "developer_handle": "some_developer_handle",
        }

        # operation
        with patch.object(
            self.strategy, "_valid_signature", side_effect=[True, False]
        ) as mock_valid:
            is_valid, code, msg = self.strategy.valid_registration(
                incorrect_registration_info, self.address
            )

        # after
        mock_valid.assert_called()

        assert not is_valid
        assert code == 1
        assert msg == "ethereum address and signature do not match!"

    def test_valid_registration_fails_v(self):
        """Test the valid_registration method of the Strategy class which fails because no developer_handle was provided."""
        # setup
        incorrect_registration_info = {
            "ethereum_address": "some_ethereum_address",
            "fetchai_address": self.address,
            "signature_of_ethereum_address": "some_signature_of_ethereum_address",
            "signature_of_fetchai_address": "some_signature_of_fetchai_address",
            "developer_handle": "",
        }

        # operation
        with patch.object(
            self.strategy, "_valid_signature", return_value=True
        ) as mock_valid:
            is_valid, code, msg = self.strategy.valid_registration(
                incorrect_registration_info, self.address
            )

        # after
        mock_valid.assert_called()

        assert not is_valid
        assert code == 1
        assert msg == "missing developer_handle!"

    def test_valid_registration_fails_vi(self):
        """Test the valid_registration method of the Strategy class which fails because agent registration is in progress."""
        # setup
        registration_info = {
            "ethereum_address": "some_ethereum_address",
            "fetchai_address": self.address,
            "signature_of_ethereum_address": "some_signature_of_ethereum_address",
            "signature_of_fetchai_address": "some_signature_of_fetchai_address",
            "developer_handle": "some_developer_handle",
        }
        self.strategy.lock_registration_temporarily(self.address, self.info)

        # operation
        with patch.object(
            self.strategy, "_valid_signature", return_value=True
        ) as mock_valid:
            is_valid, code, msg = self.strategy.valid_registration(
                registration_info, self.address
            )

        # after
        mock_valid.assert_called()

        assert not is_valid
        assert code == 1
        assert msg == "registration in process for this address!"

    def test_valid_registration_fails_vii(self):
        """Test the valid_registration method of the Strategy class which fails because agent already registered."""
        # setup
        registration_info = {
            "ethereum_address": "some_ethereum_address",
            "fetchai_address": self.address,
            "signature_of_ethereum_address": "some_signature_of_ethereum_address",
            "signature_of_fetchai_address": "some_signature_of_fetchai_address",
            "developer_handle": "some_developer_handle",
        }

        # operation
        with patch.object(
            self.strategy, "_valid_signature", return_value=True
        ) as mock_valid:
            with patch.object(self.db, "is_registered", return_value=True) as mock_is:
                is_valid, code, msg = self.strategy.valid_registration(
                    registration_info, self.address
                )

        # after
        mock_valid.assert_called()
        mock_is.assert_called_once()

        assert not is_valid
        assert code == 1
        assert msg == "already registered!"

    def test__valid_signature_i(self):
        """Test the _valid_signature method of the Strategy class where result is True."""
        # setup
        expected_signer = "some_expected_signer"
        signature = "some_signature"
        message_str = "some_message_str"
        ledger_id = "some_ledger_id"

        # operation
        with patch.object(
            LedgerApis, "recover_message", return_value=(expected_signer,)
        ) as mock_recover:
            is_valid = self.strategy._valid_signature(
                expected_signer, signature, message_str, ledger_id
            )

        # after
        mock_recover.assert_called_once()
        assert is_valid

    def test__valid_signature_ii(self):
        """Test the _valid_signature method of the Strategy class where result is False."""
        # setup
        expected_signer = "some_expected_signer"
        signature = "some_signature"
        message_str = "some_message_str"
        ledger_id = "some_ledger_id"

        # operation
        with patch.object(
            LedgerApis, "recover_message", return_value=("some_other_signer",)
        ) as mock_recover:
            is_valid = self.strategy._valid_signature(
                expected_signer, signature, message_str, ledger_id
            )

        # after
        mock_recover.assert_called_once()
        assert not is_valid

    def test__valid_signature_iii(self):
        """Test the _valid_signature method of the Strategy class where an exception is raised."""
        # setup
        expected_signer = "some_expected_signer"
        signature = "some_signature"
        message_str = "some_message_str"
        ledger_id = "some_ledger_id"

        exception_message = "some_exception_message"

        # operation
        with patch.object(
            LedgerApis, "recover_message", side_effect=Exception(exception_message)
        ) as mock_recover:
            with patch.object(self.logger, "log") as mock_logger:
                is_valid = self.strategy._valid_signature(
                    expected_signer, signature, message_str, ledger_id
                )

        # after
        mock_recover.assert_called_once()
        mock_logger.assert_any_call(
            logging.WARNING, f"Signing exception: {exception_message}",
        )
        assert not is_valid

    def test_get_terms(self):
        """Test the get_terms method of the Strategy class."""
        # setup
        counterparty = "some_counterparty"
        expected_terms = Terms(
            ledger_id=self.strategy._ledger_id,
            sender_address=self.skill.skill_context.agent_address,
            counterparty_address=counterparty,
            amount_by_currency_id={
                self.token_denomination: -self.token_dispense_amount
            },
            quantities_by_good_id={},
            is_sender_payable_tx_fee=True,
            nonce="some",
            fee_by_currency_id={self.token_denomination: self.strategy._max_tx_fee},
        )

        # operation
        actual_terms = self.strategy.get_terms(counterparty)

        # after
        assert actual_terms == expected_terms

    def test_get_kwargs(self):
        """Test the get_kwargs method of the Strategy class."""
        # setup
        expected_kwargs = {"address": self.info["ethereum_address"]}

        # operation
        actual_kwargs = self.strategy.get_kwargs(self.info)

        # after
        assert actual_kwargs == expected_kwargs

    def test_has_staked_i(self):
        """Test the get_kwargs method of the Strategy class where _override_staking_check is False and stake value is greater than 0."""
        # setup
        state = {"stake": "100"}

        # operation
        has_staked = self.strategy.has_staked(state)

        # after
        assert has_staked is True

    def test_has_staked_ii(self):
        """Test the get_kwargs method of the Strategy class where _override_staking_check is False and stake value is 0."""
        # setup
        state = {"stake": "0"}

        # operation
        has_staked = self.strategy.has_staked(state)

        # after
        assert has_staked is False

    def test_has_staked_iii(self):
        """Test the get_kwargs method of the Strategy class where _override_staking_check is True."""
        # setup
        self.strategy._override_staking_check = True
        state = {"stake": "100"}

        # operation
        has_staked = self.strategy.has_staked(state)

        # after
        assert has_staked is True



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Test P2PLibp2p connection build."""
import os
import tempfile
from io import StringIO
from unittest import mock

import pytest

from aea.exceptions import AEAException

from packages.fetchai.connections.p2p_libp2p import check_dependencies
from packages.fetchai.connections.p2p_libp2p.check_dependencies import (
    MINIMUM_GCC_VERSION,
    MINIMUM_GO_VERSION,
    build_node,
    check_versions,
    version_to_string,
)
from packages.fetchai.connections.p2p_libp2p.connection import LIBP2P_NODE_MODULE_NAME


def test_check_versions():
    """Test check_versions - positive case."""
    with mock.patch("sys.stdout", new_callable=StringIO) as mock_stdout:
        check_versions()
        stdout = mock_stdout.getvalue()
    assert f"check 'go'>={version_to_string(MINIMUM_GO_VERSION)}, found " in stdout
    assert f"check 'gcc'>={version_to_string(MINIMUM_GCC_VERSION)}, found " in stdout


def test_check_versions_negative_binary_not_found():
    """Test check_versions - negative case, binary not found."""
    with mock.patch("shutil.which", return_value=None):
        with pytest.raises(
            AEAException,
            match="'go' is required by the libp2p connection, but it is not installed, or it is not accessible from the system path.",
        ):
            check_versions()


def test_check_versions_negative_version_too_low():
    """Test check_versions - negative case, version too low."""
    with mock.patch.object(
        check_dependencies, "get_version", return_value=(0, 0, 0),
    ):
        with pytest.raises(
            AEAException,
            match=f"The installed version of 'go' is too low: expected at least {version_to_string(MINIMUM_GO_VERSION)}; found 0.0.0.",
        ):
            check_versions()


def test_check_versions_negative_cannot_parse_version():
    """Test the check_versions - negative case, cannot parse version."""
    with mock.patch("sys.stdout", new_callable=StringIO) as mock_stdout:
        with mock.patch("subprocess.check_output", return_value=b""):
            check_versions()
        stdout = mock_stdout.getvalue()
    assert (
        "Warning: cannot parse 'go' version from command: ['go', 'version']." in stdout
    )
    assert (
        "Warning: cannot parse 'gcc' version from command: ['gcc', '--version']."
        in stdout
    )


def test_build_node():
    """Test build node function."""
    with tempfile.TemporaryDirectory() as build_dir:
        build_node(build_dir)
        assert os.path.exists(os.path.join(build_dir, LIBP2P_NODE_MODULE_NAME))



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the helpers module of the tac control contract skill."""

from pathlib import Path
from unittest.mock import patch

import pytest

from aea.exceptions import AEAEnforceError
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.skills.tac_control_contract.helpers import (
    ERC1155Contract,
    _sample_good_instances,
    determine_scaling_factor,
    generate_currency_endowments,
    generate_currency_id_to_name,
    generate_currency_ids,
    generate_equilibrium_prices_and_holdings,
    generate_exchange_params,
    generate_good_endowments,
    generate_good_id_to_name,
    generate_good_ids,
    generate_utility_params,
)

from tests.conftest import ROOT_DIR


class TestHelpers(BaseSkillTestCase):
    """Test Helper module methods of tac control."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_control")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()

    def test_generate_good_ids_succeeds(self):
        """Test the generate_good_ids of Helpers module which succeeds."""
        expected_list = [1, 2, 3, 4, 5]
        with patch.object(
            ERC1155Contract, "generate_token_ids", return_value=expected_list
        ):
            good_ids = generate_good_ids(5)
        assert good_ids == expected_list

    def test_generate_good_ids_fails(self):
        """Test the generate_good_ids of Helpers module which fails because generate_token_ids generates wrong good ids."""
        expected_list = [1, 2, 3, 4, 5, 6]
        with patch.object(
            ERC1155Contract, "generate_token_ids", return_value=expected_list
        ):
            with pytest.raises(
                AEAEnforceError,
                match="Length of good ids and number of goods must match.",
            ):
                assert generate_good_ids(5)

    def test_generate_currency_ids_succeeds(self):
        """Test the generate_good_ids of Helpers module which succeeds."""
        expected_list = [1, 2, 3, 4, 5]
        with patch.object(
            ERC1155Contract, "generate_token_ids", return_value=expected_list
        ):
            currency_ids = generate_currency_ids(5)
        assert currency_ids == expected_list

    def test_generate_currency_ids_fails(self):
        """Test the generate_good_ids of Helpers module which fails because generate_token_ids generates wrong currency ids."""
        expected_list = [1, 2, 3, 4, 5, 6]
        with patch.object(
            ERC1155Contract, "generate_token_ids", return_value=expected_list
        ):
            with pytest.raises(
                AEAEnforceError,
                match="Length of currency ids and number of currencies must match.",
            ):
                assert generate_currency_ids(5)

    def test_generate_currency_id_to_name(self):
        """Test the generate_currency_id_to_name of Helpers module."""
        expected_currency_id_to_name = {
            "1": "FT_1",
            "3": "FT_3",
            "5": "FT_5",
            "7": "FT_7",
            "9": "FT_9",
        }
        currency_id_to_name = generate_currency_id_to_name([1, 3, 5, 7, 9])
        assert currency_id_to_name == expected_currency_id_to_name

    def test_generate_good_id_to_name(self):
        """Test the generate_good_id_to_name of Helpers module."""
        expected_good_id_to_name = {
            "1": "FT_1",
            "3": "FT_3",
            "5": "FT_5",
            "7": "FT_7",
            "9": "FT_9",
        }
        good_id_to_name = generate_good_id_to_name([1, 3, 5, 7, 9])
        assert good_id_to_name == expected_good_id_to_name

    def test_determine_scaling_factor(self):
        """Test the determine_scaling_factor of Helpers module."""
        money_endowment = 53730411
        scaling_factor = determine_scaling_factor(money_endowment)
        assert scaling_factor == 10000000.0

    def test_generate_good_endowments(self):
        """Test the generate_good_endowments of Helpers module."""
        endowments = generate_good_endowments(
            ["ag_1_add", "ag_2_add"], ["good_id_1", "good_id_2"], 2, 1, 1
        )
        assert "good_id_1" in endowments["ag_1_add"]
        assert "good_id_2" in endowments["ag_1_add"]
        assert "good_id_1" in endowments["ag_2_add"]
        assert "good_id_2" in endowments["ag_2_add"]

    def test_generate_utility_params(self):
        """Test the generate_utility_params of Helpers module."""
        utility_function_params = generate_utility_params(
            ["ag_1_add", "ag_2_add"], ["good_id_1", "good_id_2"], 1000.0
        )
        assert "good_id_1" in utility_function_params["ag_1_add"].keys()
        assert "good_id_2" in utility_function_params["ag_1_add"].keys()
        assert "good_id_1" in utility_function_params["ag_2_add"].keys()
        assert "good_id_2" in utility_function_params["ag_2_add"].keys()

    def test_sample_good_instances(self):
        """Test the _sample_good_instances of Helpers module."""
        nb_instances = _sample_good_instances(2, ["good_id_1", "good_id_2"], 2, 1, 1)
        assert type(nb_instances["good_id_1"]) == int
        assert type(nb_instances["good_id_2"]) == int

    def test_generate_currency_endowments(self):
        """Test the generate_currency_endowments of Helpers module."""
        currency_endowments = generate_currency_endowments(
            ["ag_1_add", "ag_2_add"], ["currency_id_1", "currency_id_2"], 10
        )
        assert "currency_id_1" in currency_endowments["ag_1_add"].keys()
        assert "currency_id_2" in currency_endowments["ag_2_add"].keys()
        assert currency_endowments["ag_1_add"]["currency_id_1"] == 10
        assert currency_endowments["ag_1_add"]["currency_id_2"] == 10

        assert currency_endowments["ag_2_add"]["currency_id_1"] == 10
        assert currency_endowments["ag_2_add"]["currency_id_2"] == 10

    def test_generate_exchange_params(self):
        """Test the generate_exchange_params of Helpers module."""
        currency_endowments = generate_exchange_params(
            ["ag_1_add", "ag_2_add"], ["currency_id_1", "currency_id_2"]
        )
        assert "currency_id_1" in currency_endowments["ag_1_add"].keys()
        assert "currency_id_2" in currency_endowments["ag_2_add"].keys()
        assert currency_endowments["ag_1_add"]["currency_id_1"] == 1.0
        assert currency_endowments["ag_1_add"]["currency_id_2"] == 1.0

        assert currency_endowments["ag_2_add"]["currency_id_1"] == 1.0
        assert currency_endowments["ag_2_add"]["currency_id_2"] == 1.0

    def test_generate_equilibrium_prices_and_holdings(self):
        """Test the generate_equilibrium_prices_and_holdings of Helpers module."""
        (
            eq_prices_dict,
            eq_good_holdings_dict,
            eq_currency_holdings_dict,
        ) = generate_equilibrium_prices_and_holdings(
            {"ag_1": {"good_1": 1}},
            {"ag_1": {"good_1": 1.0}},
            {"ag_1": {"currency_1": 10}},
            {"ag_1": {"currency_1": 1.0}},
            2.0,
        )

        assert len(eq_prices_dict) == 1
        assert type(eq_prices_dict["good_1"]) == float

        assert len(eq_good_holdings_dict) == 1
        assert len(eq_good_holdings_dict["ag_1"]) == 1
        assert type(eq_good_holdings_dict["ag_1"]["good_1"]) == float

        assert len(eq_currency_holdings_dict) == 1
        assert type(eq_currency_holdings_dict["ag_1"]) == float



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains types and helpers for acn Proof-of-Representation."""
from typing import Optional

from aea.common import PathLike
from aea.crypto.registries import make_ledger_api
from aea.helpers.base import CertRequest


class AgentRecord:
    """Agent Proof-of-Representation to representative."""

    __slots__ = (
        "_address",
        "_representative_public_key",
        "_message",
        "_signature",
        "_ledger_id",
        "_public_key",
    )

    def __init__(
        self,
        address: str,
        representative_public_key: str,
        message: bytes,
        signature: str,
        ledger_id: str,
    ) -> None:
        """
        Initialize the AgentRecord

        :param address: agent address
        :param representative_public_key: representative's public key
        :param message: message to be signed as proof-of-represenation of this AgentRecord
        :param signature: proof-of-representation of this AgentRecord
        :param ledger_id: ledger id
        """
        self._address = address
        self._representative_public_key = representative_public_key
        self._message = message
        self._signature = signature
        self._ledger_id = ledger_id
        self._public_key: Optional[str] = None
        self._check_validity()

    def _check_validity(self) -> None:
        """
        Checks validity of record.

        Specificyally:
        - if ledger_id is valid
        - if agent signed the message
        - if message is correctly formatted
        """
        if self.message != self._get_message(self.representative_public_key):
            raise ValueError("Invalid message.")  # pragma: no cover
        ledger_api = make_ledger_api(self.ledger_id)
        public_keys = ledger_api.recover_public_keys_from_message(
            self.message, self.signature
        )
        if len(public_keys) == 0:
            raise ValueError("Malformed signature!")  # pragma: no cover
        public_key: Optional[str] = None
        for public_key_ in public_keys:
            address = ledger_api.get_address_from_public_key(public_key_)
            if address == self.address:
                public_key = public_key_
                break
        if public_key is None:
            raise ValueError(
                "Invalid signature for provided representative_public_key and agent address!"
            )
        self._public_key = public_key

    @property
    def address(self) -> str:
        """Get agent address"""
        return self._address

    @property
    def public_key(self) -> str:
        """Get agent public key"""
        if self._public_key is None:
            raise ValueError("Inconsistent record!")  #  pragma: nocover
        return self._public_key

    @property
    def representative_public_key(self) -> str:
        """Get agent representative's public key"""
        return self._representative_public_key

    @property
    def signature(self) -> str:
        """Get record signature"""
        return self._signature

    @property
    def message(self) -> bytes:
        """Get the message."""
        return self._message

    @property
    def ledger_id(self) -> str:
        """Get ledger id."""
        return self._ledger_id

    def _get_message(self, public_key: str) -> bytes:  # pylint: disable=no-self-use
        """Get the message."""
        # Refactor, needs to match CertRequest!
        message = public_key.encode("ascii")
        # + self.identifier.encode("ascii")  # noqa: E800
        # + self.not_before_string.encode("ascii")  # noqa: E800
        # + self.not_after_string.encode("ascii")  # noqa: E800
        return message

    def __str__(self) -> str:  # pragma: no cover
        """Get string representation."""
        return f"(address={self.address}, public_key={self.public_key}, representative_public_key={self.representative_public_key}, signature={self.signature}, ledger_id={self.ledger_id})"

    @classmethod
    def from_cert_request(
        cls,
        cert_request: CertRequest,
        address: str,
        representative_public_key: str,
        data_dir: Optional[PathLike] = None,
    ) -> "AgentRecord":
        """Get agent record from cert request."""
        message = cert_request.get_message(representative_public_key)
        signature = cert_request.get_signature(data_dir)
        record = cls(
            address,
            representative_public_key,
            message,
            signature,
            cert_request.ledger_id,
        )
        return record
